from graphene_django.utils.testing import GraphQLTestCase
from unittest.mock import patch

from django.contrib.auth.models import Group
from django.test import override_settings

from allianceauth.tests.test_auth_utils import AuthUtils
from app_utils.testdata_factories import UserFactory
from app_utils.testing import generate_invalid_pk

from allianceauth.groupmanagement.models import GroupRequest, RequestLog

from ..groupmanagement.types import GroupRequestLogType, GroupRequestLogActionType, GroupRequestAddStatus, GroupRequestLeaveStatus
from ..authentication.types import GroupStatusEnum


class TestQueries(GraphQLTestCase):
    maxDiff = None

    @classmethod
    def setUpTestData(cls):
        cls.user = AuthUtils.add_permission_to_user_by_name('auth.group_management', UserFactory(), False)

        cls.user2 = UserFactory()

        cls.group1 = Group.objects.create(name="Test Group 1")

        cls.authgroup1 = cls.group1.authgroup

        cls.authgroup1.internal = False
        cls.authgroup1.open = False
        cls.authgroup1.hidden = False
        cls.authgroup1.public = True
        cls.authgroup1.save()

        cls.group2 = Group.objects.create(name="Test Group 2")

        cls.authgroup2 = cls.group2.authgroup

        cls.authgroup2.internal = False
        cls.authgroup2.open = False
        cls.authgroup2.hidden = False
        cls.authgroup2.public = True
        cls.authgroup2.save()

        cls.user2.groups.add(cls.group2)
        cls.authgroup2.group_leaders.add(cls.user2)

        GroupRequest.objects.create(
            user=cls.user,
            group=cls.group1,
            leave_request=False,
        )

        cls.user3 = UserFactory()
        cls.user3.groups.add(cls.group2)

        GroupRequest.objects.create(
            user=cls.user3,
            group=cls.group2,
            leave_request=True,
        )

        cls.log1 = RequestLog.objects.create(
            request_type=None,
            group=cls.group1,
            request_info=f'{cls.user2.username}:{cls.group1.name}',
            action=False,
            request_actor=cls.user,
        )

        cls.log2 = RequestLog.objects.create(
            request_type=True,
            group=cls.group1,
            request_info=f'{cls.user2.username}:{cls.group1.name}',
            action=True,
            request_actor=cls.user,
        )

        cls.log3 = RequestLog.objects.create(
            request_type=False,
            group=cls.group1,
            request_info=f'{cls.user2.username}:{cls.group1.name}',
            action=False,
            request_actor=cls.user,
        )

    def test_user_joinable_groups(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            query {
                groupmanagementUserJoinableGroups {
                    id
                    status
                }
            }
            '''
        )

        self.assertJSONEqual(
            response.content,
            {
                "data": {
                    "groupmanagementUserJoinableGroups": [
                        {
                            "id": str(self.group1.id),
                            "status": GroupStatusEnum.PENDING.name,
                        },
                        {
                            "id": str(self.group2.id),
                            "status": GroupStatusEnum.CAN_APPLY.name,
                        },
                    ]
                }
            }
        )

    def test_group_management_has_perms(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            query {
                groupmanagementManageRequests {
                    leaveRequests {
                        group {
                            id
                        }
                        user {
                            id
                        }
                    }
                    acceptRequests {
                        group {
                            id
                        }
                        user {
                            id
                        }
                    }
                }
            }
            '''
        )

        self.assertJSONEqual(
            response.content,
            {
                "data": {
                    "groupmanagementManageRequests": {
                        "leaveRequests": [
                            {
                                "group": {
                                    "id": str(self.group2.id),
                                },
                                "user": {
                                    "id": str(self.user3.id),
                                }
                            }
                        ],
                        "acceptRequests": [
                            {
                                "group": {
                                    "id": str(self.group1.id),
                                },
                                "user": {
                                    "id": str(self.user.id),
                                }
                            }
                        ]
                    }
                }
            }
        )

    def test_group_management_group_leader(self):
        self.client.force_login(self.user2)

        response = self.query(
            '''
            query {
                groupmanagementManageRequests {
                    leaveRequests {
                        group {
                            id
                        }
                        user {
                            id
                        }
                    }
                    acceptRequests {
                        group {
                            id
                        }
                        user {
                            id
                        }
                    }
                }
            }
            '''
        )

        self.assertJSONEqual(
            response.content,
            {
                "data": {
                    "groupmanagementManageRequests": {
                        "leaveRequests": [
                            {
                                "group": {
                                    "id": str(self.group2.id),
                                },
                                "user": {
                                    "id": str(self.user3.id),
                                }
                            }
                        ],
                        "acceptRequests": []
                    }
                }
            }
        )

    def test_group_membership_has_perms(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            query {
                groupmanagementGroups {
                    id
                    numMembers
                }
            }
            '''
        )

        self.assertJSONEqual(
            response.content,
            {
                "data": {
                    "groupmanagementGroups": [
                        {
                            "id": str(self.group1.id),
                            "numMembers": 0,
                        },
                        {
                            "id": str(self.group2.id),
                            "numMembers": 2,
                        },
                    ]
                }
            }
        )

    def test_group_membership_group_leader(self):
        self.client.force_login(self.user2)

        response = self.query(
            '''
            query {
                groupmanagementGroups {
                    id
                    numMembers
                }
            }
            '''
        )

        self.assertJSONEqual(
            response.content,
            {
                "data": {
                    "groupmanagementGroups": [
                        {
                            'id': str(self.group2.id),
                            'numMembers': 2,
                        }
                    ]
                }
            }
        )

    def test_group_membership_list_ok(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            query($groupId: Int!) {
                groupmanagementGroupMemberships(groupId: $groupId) {
                    group {
                        id
                    }
                    members {
                        user {
                            id
                        }
                        isLeader
                    }
                }
            }
            ''',
            variables={
                "groupId": self.group2.id,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                "data": {
                    "groupmanagementGroupMemberships": {
                        'group': {
                            'id': str(self.group2.id),
                        },
                        "members": [
                            {
                                "user": {
                                    "id": str(user.id),
                                },
                                "isLeader": user == self.user2,
                            } for user in [self.user2, self.user3]
                        ]
                    }
                }
            }
        )

    def test_group_membership_list_permission_denied(self):
        self.client.force_login(self.user2)

        response = self.query(
            '''
            query($groupId: Int!) {
                groupmanagementGroupMemberships(groupId: $groupId) {
                    group {
                        id
                    }
                    members {
                        user {
                            id
                        }
                        isLeader
                    }
                }
            }
            ''',
            variables={
                "groupId": self.group1.id,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                "data": {
                    "groupmanagementGroupMemberships": None
                },
                "errors": [
                    {
                        "message": 'An unknown error occurred.',
                        "locations": [
                            {
                                "line": 3,
                                "column": 17
                            }
                        ],
                        "path": [
                            "groupmanagementGroupMemberships"
                        ]
                    }
                ]
            }
        )

    def test_group_membership_list_not_found(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            query($groupId: Int!) {
                groupmanagementGroupMemberships(groupId: $groupId) {
                    group {
                        id
                    }
                    members {
                        user {
                            id
                        }
                        isLeader
                    }
                }
            }
            ''',
            variables={
                "groupId": generate_invalid_pk(Group),
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                "data": {
                    "groupmanagementGroupMemberships": None
                },
                "errors": [
                    {
                        "message": "Group doesn't exist",
                        "locations": [
                            {
                                "line": 3,
                                "column": 17
                            }
                        ],
                        "path": [
                            "groupmanagementGroupMemberships"
                        ]
                    }
                ]
            }
        )

    def test_group_membership_audit_ok(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            query($groupId: Int!) {
                groupmanagementGroupMembershipAudit(groupId: $groupId) {
                    group {
                        id
                    }
                    entries {
                        requestType
                        action
                        requestor {
                            id
                        }
                    }
                }
            }
            ''',
            variables={
                "groupId": self.group1.id,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                "data": {
                    "groupmanagementGroupMembershipAudit": {
                        'group': {
                            'id': str(self.group1.id),
                        },
                        "entries": [
                            {
                                "requestType": GroupRequestLogType.JOIN.name,
                                "action": GroupRequestLogActionType.REJECT.name,
                                "requestor": {
                                    "id": str(self.user2.id),
                                }
                            },
                            {
                                "requestType": GroupRequestLogType.LEAVE.name,
                                "action": GroupRequestLogActionType.ACCEPT.name,
                                "requestor": {
                                    "id": str(self.user2.id),
                                }
                            },
                            {
                                "requestType": GroupRequestLogType.REMOVED.name,
                                "action": None,
                                "requestor": {
                                    "id": str(self.user2.id),
                                }
                            },
                        ]
                    }
                }
            }
        )

    def test_group_membership_audit_permission_denied(self):
        self.client.force_login(self.user2)

        response = self.query(
            '''
            query($groupId: Int!) {
                groupmanagementGroupMembershipAudit(groupId: $groupId) {
                    group {
                        id
                    }
                    entries {
                        requestType
                        action
                        requestor {
                            id
                        }
                    }
                }
            }
            ''',
            variables={
                "groupId": self.group1.id,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                "data": {
                    "groupmanagementGroupMembershipAudit": None
                },
                "errors": [
                    {
                        "message": 'An unknown error occurred.',
                        "locations": [
                            {
                                "line": 3,
                                "column": 17
                            }
                        ],
                        "path": [
                            "groupmanagementGroupMembershipAudit"
                        ]
                    }
                ]
            }
        )

    def test_group_membership_audit_not_found(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            query($groupId: Int!) {
                groupmanagementGroupMembershipAudit(groupId: $groupId) {
                    group {
                        id
                    }
                    entries {
                        requestType
                        action
                        requestor {
                            id
                        }
                    }
                }
            }
            ''',
            variables={
                "groupId": generate_invalid_pk(Group),
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                "data": {
                    "groupmanagementGroupMembershipAudit": None
                },
                "errors": [
                    {
                        "message": "Group does not exist",
                        "locations": [
                            {
                                "line": 3,
                                "column": 17
                            }
                        ],
                        "path": [
                            "groupmanagementGroupMembershipAudit"
                        ]
                    }
                ]
            }
        )


class TestAddGroupRequestMutation(GraphQLTestCase):
    maxDiff = None

    @classmethod
    def setUpTestData(cls):
        cls.user = UserFactory()

        cls.group = Group.objects.create(name="Test Group 1")

        cls.authgroup = cls.group.authgroup

        cls.authgroup.internal = False
        cls.authgroup.open = False
        cls.authgroup.hidden = False
        cls.authgroup.public = True
        cls.authgroup.save()

        cls.authgroup.states.add(cls.user.profile.state)

    def test_ok(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            mutation($groupId: Int!) {
                groupmanagementJoinGroupRequest(groupId: $groupId) {
                    ok
                    status
                }
            }
            ''',
            variables={
                "groupId": self.group.id,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                "data": {
                    "groupmanagementJoinGroupRequest": {
                        "ok": True,
                        "status": GroupRequestAddStatus.APPLIED.name,
                    }
                }
            }
        )

        self.assertEqual(GroupRequest.objects.count(), 1)

    def test_ok_open(self):
        self.client.force_login(self.user)

        self.authgroup.open = True
        self.authgroup.save()

        response = self.query(
            '''
            mutation($groupId: Int!) {
                groupmanagementJoinGroupRequest(groupId: $groupId) {
                    ok
                    status
                }
            }
            ''',
            variables={
                "groupId": self.group.id,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                "data": {
                    "groupmanagementJoinGroupRequest": {
                        "ok": True,
                        "status": GroupRequestAddStatus.JOINED.name,
                    }
                }
            }
        )

        self.assertIn(self.group, self.user.groups.all())
        self.assertEqual(RequestLog.objects.count(), 1)

    def test_not_joinable(self):
        self.client.force_login(self.user)

        self.authgroup.states.clear()
        self.authgroup.internal = True
        self.authgroup.save()

        response = self.query(
            '''
            mutation($groupId: Int!) {
                groupmanagementJoinGroupRequest(groupId: $groupId) {
                    ok
                    status
                }
            }
            ''',
            variables={
                "groupId": self.group.id,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                "data": {
                    "groupmanagementJoinGroupRequest": {
                        "ok": False,
                        "status": GroupRequestAddStatus.CANNOT_JOIN.name,
                    }
                }
            }
        )

        self.assertEqual(GroupRequest.objects.count(), 0)

    def test_already_in_group(self):
        self.client.force_login(self.user)

        self.user.groups.add(self.group)

        response = self.query(
            '''
            mutation($groupId: Int!) {
                groupmanagementJoinGroupRequest(groupId: $groupId) {
                    ok
                    status
                }
            }
            ''',
            variables={
                "groupId": self.group.id,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                "data": {
                    "groupmanagementJoinGroupRequest": {
                        "ok": False,
                        "status": GroupRequestAddStatus.ALREADY_MEMBER.name,
                    }
                }
            }
        )

        self.assertEqual(GroupRequest.objects.count(), 0)

    def test_group_not_public_and_no_perms(self):
        self.client.force_login(self.user)

        self.authgroup.public = False
        self.authgroup.save()

        response = self.query(
            '''
            mutation($groupId: Int!) {
                groupmanagementJoinGroupRequest(groupId: $groupId) {
                    ok
                    status
                }
            }
            ''',
            variables={
                "groupId": self.group.id,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                "data": {
                    "groupmanagementJoinGroupRequest": {
                        "ok": False,
                        "status": GroupRequestAddStatus.CANNOT_JOIN.name,
                    }
                }
            }
        )

        self.assertEqual(GroupRequest.objects.count(), 0)

    def test_already_applied(self):
        self.client.force_login(self.user)

        GroupRequest.objects.create(
            group=self.group,
            user=self.user,
        )

        response = self.query(
            '''
            mutation($groupId: Int!) {
                groupmanagementJoinGroupRequest(groupId: $groupId) {
                    ok
                    status
                }
            }
            ''',
            variables={
                "groupId": self.group.id,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                "data": {
                    "groupmanagementJoinGroupRequest": {
                        "ok": False,
                        "status": GroupRequestAddStatus.ALREADY_APPLIED.name,
                    }
                }
            }
        )

        self.assertEqual(GroupRequest.objects.count(), 1)


class TestLeaveGroupRequestMutation(GraphQLTestCase):
    maxDiff = None

    @classmethod
    def setUpTestData(cls):
        cls.user = UserFactory()

        cls.group = Group.objects.create(name="Test Group 1")

        cls.authgroup = cls.group.authgroup

        cls.authgroup.internal = False
        cls.authgroup.open = False
        cls.authgroup.hidden = False
        cls.authgroup.public = True
        cls.authgroup.save()

        cls.user.groups.add(cls.group)

    def test_ok(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            mutation($groupId: Int!) {
                groupmanagementLeaveGroupRequest(groupId: $groupId) {
                    ok
                    status
                }
            }
            ''',
            variables={
                "groupId": self.group.id,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                "data": {
                    "groupmanagementLeaveGroupRequest": {
                        "ok": True,
                        "status": GroupRequestLeaveStatus.CREATED_LEAVE_REQUEST.name,
                    }
                }
            }
        )

        self.assertEqual(GroupRequest.objects.count(), 1)

    def test_ok_open(self):
        self.client.force_login(self.user)

        self.authgroup.open = True
        self.authgroup.save()

        response = self.query(
            '''
            mutation($groupId: Int!) {
                groupmanagementLeaveGroupRequest(groupId: $groupId) {
                    ok
                    status
                }
            }
            ''',
            variables={
                "groupId": self.group.id,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                "data": {
                    "groupmanagementLeaveGroupRequest": {
                        "ok": True,
                        "status": GroupRequestLeaveStatus.LEFT.name,
                    }
                }
            }
        )

        self.assertNotIn(self.group, self.user.groups.all())
        self.assertEqual(RequestLog.objects.count(), 1)

    def test_internal(self):
        self.client.force_login(self.user)

        self.authgroup.internal = True
        self.authgroup.save()

        response = self.query(
            '''
            mutation($groupId: Int!) {
                groupmanagementLeaveGroupRequest(groupId: $groupId) {
                    ok
                    status
                }
            }
            ''',
            variables={
                "groupId": self.group.id,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                "data": {
                    "groupmanagementLeaveGroupRequest": {
                        "ok": False,
                        "status": GroupRequestLeaveStatus.CANNOT_LEAVE.name,
                    }
                }
            }
        )

        self.assertEqual(GroupRequest.objects.count(), 0)

    def test_not_in_group(self):
        self.client.force_login(self.user)

        self.user.groups.remove(self.group)

        response = self.query(
            '''
            mutation($groupId: Int!) {
                groupmanagementLeaveGroupRequest(groupId: $groupId) {
                    ok
                    status
                }
            }
            ''',
            variables={
                "groupId": self.group.id,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                "data": {
                    "groupmanagementLeaveGroupRequest": {
                        "ok": False,
                        "status": GroupRequestLeaveStatus.NOT_MEMBER.name,
                    }
                }
            }
        )

        self.assertEqual(GroupRequest.objects.count(), 0)

    def test_already_applied(self):
        self.client.force_login(self.user)

        GroupRequest.objects.create(
            group=self.group,
            user=self.user,
        )

        response = self.query(
            '''
            mutation($groupId: Int!) {
                groupmanagementLeaveGroupRequest(groupId: $groupId) {
                    ok
                    status
                }
            }
            ''',
            variables={
                "groupId": self.group.id,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                "data": {
                    "groupmanagementLeaveGroupRequest": {
                        "ok": False,
                        "status": GroupRequestLeaveStatus.PENDING_LEAVE_REQUEST.name,
                    }
                }
            }
        )

        self.assertEqual(GroupRequest.objects.count(), 1)

    @override_settings(GROUPMANAGEMENT_AUTO_LEAVE=True)
    def test_auto_leave(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            mutation($groupId: Int!) {
                groupmanagementLeaveGroupRequest(groupId: $groupId) {
                    ok
                    status
                }
            }
            ''',
            variables={
                "groupId": self.group.id,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                "data": {
                    "groupmanagementLeaveGroupRequest": {
                        "ok": True,
                        "status": GroupRequestLeaveStatus.LEFT.name,
                    }
                }
            }
        )

        self.assertNotIn(self.group, self.user.groups.all())
        self.assertEqual(RequestLog.objects.count(), 1)


class TestGroupMembershipRemove(GraphQLTestCase):
    maxDiff = None

    @classmethod
    def setUpTestData(cls):
        cls.user = UserFactory()
        cls.user2 = UserFactory()

        cls.group = Group.objects.create(name="Test Group 1")

        cls.authgroup = cls.group.authgroup

        cls.authgroup.internal = False
        cls.authgroup.open = False
        cls.authgroup.hidden = False
        cls.authgroup.public = True
        cls.authgroup.save()

        cls.user2.groups.add(cls.group)

        cls.authgroup.group_leaders.add(cls.user)

    def test_ok(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            mutation($groupId: Int!, $userId: Int!) {
                groupmanagementRemoveMember(groupId: $groupId, userId: $userId) {
                    ok
                    error
                }
            }
            ''',
            variables={
                "groupId": self.group.id,
                "userId": self.user2.id,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                "data": {
                    "groupmanagementRemoveMember": {
                        "ok": True,
                        "error": None,
                    }
                }
            }
        )

        self.assertNotIn(self.group, self.user2.groups.all())
        self.assertEqual(RequestLog.objects.count(), 1)

    def test_cannot_manage(self):
        self.client.force_login(self.user2)

        group2 = Group.objects.create(name="Test Group 2")
        group2.authgroup.group_leaders.add(self.user2)

        response = self.query(
            '''
            mutation($groupId: Int!, $userId: Int!) {
                groupmanagementRemoveMember(groupId: $groupId, userId: $userId) {
                    ok
                    error
                }
            }
            ''',
            variables={
                "groupId": self.group.id,
                "userId": self.user2.id,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                "data": {
                    "groupmanagementRemoveMember": {
                        "ok": False,
                        "error": 'Permission denied',
                    }
                }
            }
        )

        self.assertIn(self.group, self.user2.groups.all())

    def test_not_in_group(self):
        self.client.force_login(self.user)

        self.user2.groups.remove(self.group)

        response = self.query(
            '''
            mutation($groupId: Int!, $userId: Int!) {
                groupmanagementRemoveMember(groupId: $groupId, userId: $userId) {
                    ok
                    error
                }
            }
            ''',
            variables={
                "groupId": self.group.id,
                "userId": self.user2.id,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                "data": {
                    "groupmanagementRemoveMember": {
                        "ok": False,
                        "error": "User does not exist in that group",
                    }
                }
            }
        )

        self.assertNotIn(self.group, self.user2.groups.all())

    def test_group_not_exists(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            mutation($groupId: Int!, $userId: Int!) {
                groupmanagementRemoveMember(groupId: $groupId, userId: $userId) {
                    ok
                    error
                }
            }
            ''',
            variables={
                "groupId": generate_invalid_pk(Group),
                "userId": self.user2.id,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                "data": {
                    "groupmanagementRemoveMember": {
                        "ok": False,
                        "error": "Group does not exist",
                    }
                }
            }
        )


class TestGroupMembershipAcceptAndRejectRequestMutations(GraphQLTestCase):
    maxDiff = None

    @classmethod
    def setUpTestData(cls):
        cls.user = UserFactory()
        cls.user2 = UserFactory()

        cls.group = Group.objects.create(name="Test Group 1")

        cls.authgroup = cls.group.authgroup

        cls.authgroup.internal = False
        cls.authgroup.open = False
        cls.authgroup.hidden = False
        cls.authgroup.public = True
        cls.authgroup.save()

        cls.authgroup.group_leaders.add(cls.user)

        cls.request = GroupRequest.objects.create(
            group=cls.group,
            user=cls.user2,
            leave_request=False,
        )

    def test_accept_ok(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            mutation($groupRequestId: Int!) {
                groupmanagementAcceptJoinRequest(groupRequestId: $groupRequestId) {
                    ok
                    error
                }
            }
            ''',
            variables={
                "groupRequestId": self.request.id,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                "data": {
                    "groupmanagementAcceptJoinRequest": {
                        "ok": True,
                        "error": None,
                    }
                }
            }
        )

        self.assertIn(self.group, self.user2.groups.all())
        self.assertEqual(RequestLog.objects.count(), 1)

    def test_accept_request_not_exists(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            mutation($groupRequestId: Int!) {
                groupmanagementAcceptJoinRequest(groupRequestId: $groupRequestId) {
                    ok
                    error
                }
            }
            ''',
            variables={
                "groupRequestId": generate_invalid_pk(GroupRequest),
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                "data": {
                    "groupmanagementAcceptJoinRequest": {
                        "ok": False,
                        "error": "Group request doesn't exist",
                    }
                }
            }
        )

    def test_accept_cannot_manage(self):
        self.client.force_login(self.user2)

        group2 = Group.objects.create(name="Test Group 2")
        group2.authgroup.group_leaders.add(self.user2)

        response = self.query(
            '''
            mutation($groupRequestId: Int!) {
                groupmanagementAcceptJoinRequest(groupRequestId: $groupRequestId) {
                    ok
                    error
                }
            }
            ''',
            variables={
                "groupRequestId": self.request.id,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                "data": {
                    "groupmanagementAcceptJoinRequest": {
                        "ok": False,
                        "error": 'Permission denied',
                    }
                }
            }
        )

        self.assertNotIn(self.group, self.user2.groups.all())
        self.assertEqual(RequestLog.objects.count(), 0)
        self.assertEqual(GroupRequest.objects.count(), 1)

    @patch.object(GroupRequest, 'delete')
    def test_accept_random_error(self, mock_delete):
        self.client.force_login(self.user)

        mock_delete.side_effect = Exception("Test")

        response = self.query(
            '''
            mutation($groupRequestId: Int!) {
                groupmanagementAcceptJoinRequest(groupRequestId: $groupRequestId) {
                    ok
                    error
                }
            }
            ''',
            variables={
                "groupRequestId": self.request.id,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                "data": {
                    "groupmanagementAcceptJoinRequest": {
                        "ok": False,
                        "error": f'An unhandled error occurred while processing the application from {self.request.main_char} to {self.request.group}.',
                    }
                }
            }
        )

        # Some of these asserts will fail due to delete being after adding group and log
        # I'll push a MR to AA adding a database transaction in the future

        # self.assertNotIn(self.group, self.user2.groups.all())
        # self.assertEqual(RequestLog.objects.count(), 0)
        # self.assertEqual(GroupRequest.objects.count(), 1)

    def test_reject_ok(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            mutation($groupRequestId: Int!) {
                groupmanagementRejectJoinRequest(groupRequestId: $groupRequestId) {
                    ok
                    error
                }
            }
            ''',
            variables={
                "groupRequestId": self.request.id,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                "data": {
                    "groupmanagementRejectJoinRequest": {
                        "ok": True,
                        "error": None,
                    }
                }
            }
        )

        self.assertNotIn(self.group, self.user2.groups.all())
        self.assertEqual(RequestLog.objects.count(), 1)
        self.assertEqual(GroupRequest.objects.count(), 0)

    def test_reject_request_not_exists(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            mutation($groupRequestId: Int!) {
                groupmanagementRejectJoinRequest(groupRequestId: $groupRequestId) {
                    ok
                    error
                }
            }
            ''',
            variables={
                "groupRequestId": generate_invalid_pk(GroupRequest),
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                "data": {
                    "groupmanagementRejectJoinRequest": {
                        "ok": False,
                        "error": "Group request doesn't exist",
                    }
                }
            }
        )

    def test_reject_cannot_manage(self):
        self.client.force_login(self.user2)

        group2 = Group.objects.create(name="Test Group 2")
        group2.authgroup.group_leaders.add(self.user2)

        response = self.query(
            '''
            mutation($groupRequestId: Int!) {
                groupmanagementRejectJoinRequest(groupRequestId: $groupRequestId) {
                    ok
                    error
                }
            }
            ''',
            variables={
                "groupRequestId": self.request.id,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                "data": {
                    "groupmanagementRejectJoinRequest": {
                        "ok": False,
                        "error": 'Permission denied',
                    }
                }
            }
        )

        self.assertNotIn(self.group, self.user2.groups.all())
        self.assertEqual(RequestLog.objects.count(), 0)
        self.assertEqual(GroupRequest.objects.count(), 1)

    @patch.object(GroupRequest, 'delete')
    def test_reject_random_error(self, mock_delete):
        self.client.force_login(self.user)

        mock_delete.side_effect = Exception("Test")

        response = self.query(
            '''
            mutation($groupRequestId: Int!) {
                groupmanagementRejectJoinRequest(groupRequestId: $groupRequestId) {
                    ok
                    error
                }
            }
            ''',
            variables={
                "groupRequestId": self.request.id,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                "data": {
                    "groupmanagementRejectJoinRequest": {
                        "ok": False,
                        "error": f'An unhandled error occurred while processing the application from {self.request.main_char} to {self.request.group}.',
                    }
                }
            }
        )

        # Same as test_accept_random_error

        # self.assertNotIn(self.group, self.user2.groups.all())
        # self.assertEqual(RequestLog.objects.count(), 0)
        # self.assertEqual(GroupRequest.objects.count(), 1)


class TestGroupLeaveAcceptAndRejectRequestMutation(GraphQLTestCase):
    maxDiff = None

    @classmethod
    def setUpTestData(cls):
        cls.user = UserFactory()
        cls.user2 = UserFactory()

        cls.group = Group.objects.create(name="Test Group 1")

        cls.authgroup = cls.group.authgroup

        cls.authgroup.internal = False
        cls.authgroup.open = False
        cls.authgroup.hidden = False
        cls.authgroup.public = True
        cls.authgroup.save()

        cls.authgroup.group_leaders.add(cls.user)

        cls.user2.groups.add(cls.group)

        cls.request = GroupRequest.objects.create(
            group=cls.group,
            user=cls.user2,
            leave_request=True,
        )

    def test_accept_ok(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            mutation($groupRequestId: Int!) {
                groupmanagementAcceptLeaveRequest(groupRequestId: $groupRequestId) {
                    ok
                    error
                }
            }
            ''',
            variables={
                "groupRequestId": self.request.id,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                "data": {
                    "groupmanagementAcceptLeaveRequest": {
                        "ok": True,
                        "error": None,
                    }
                }
            }
        )

        self.assertNotIn(self.group, self.user2.groups.all())
        self.assertEqual(RequestLog.objects.count(), 1)
        self.assertEqual(GroupRequest.objects.count(), 0)

    def test_accept_request_not_exists(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            mutation($groupRequestId: Int!) {
                groupmanagementAcceptLeaveRequest(groupRequestId: $groupRequestId) {
                    ok
                    error
                }
            }
            ''',
            variables={
                "groupRequestId": generate_invalid_pk(GroupRequest),
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                "data": {
                    "groupmanagementAcceptLeaveRequest": {
                        "ok": False,
                        "error": "Group request doesn't exist",
                    }
                }
            }
        )

    def test_accept_cannot_manage(self):
        self.client.force_login(self.user2)

        group2 = Group.objects.create(name="Test Group 2")
        group2.authgroup.group_leaders.add(self.user2)

        response = self.query(
            '''
            mutation($groupRequestId: Int!) {
                groupmanagementAcceptLeaveRequest(groupRequestId: $groupRequestId) {
                    ok
                    error
                }
            }
            ''',
            variables={
                "groupRequestId": self.request.id,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                "data": {
                    "groupmanagementAcceptLeaveRequest": {
                        "ok": False,
                        "error": 'Permission denied',
                    }
                }
            }
        )

        self.assertIn(self.group, self.user2.groups.all())
        self.assertEqual(RequestLog.objects.count(), 0)
        self.assertEqual(GroupRequest.objects.count(), 1)

    @patch.object(GroupRequest, 'delete')
    def test_accept_random_error(self, mock_delete):
        self.client.force_login(self.user)

        mock_delete.side_effect = Exception("Test")

        response = self.query(
            '''
            mutation($groupRequestId: Int!) {
                groupmanagementAcceptLeaveRequest(groupRequestId: $groupRequestId) {
                    ok
                    error
                }
            }
            ''',
            variables={
                "groupRequestId": self.request.id,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                "data": {
                    "groupmanagementAcceptLeaveRequest": {
                        "ok": False,
                        "error": f'An unhandled error occurred while processing the application from {self.request.main_char} to {self.request.group}.',
                    }
                }
            }
        )

        # Same as test_accept_random_error

        # self.assertIn(self.group, self.user2.groups.all())
        # self.assertEqual(RequestLog.objects.count(), 0)
        # self.assertEqual(GroupRequest.objects.count(), 1)

    def test_reject_ok(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            mutation($groupRequestId: Int!) {
                groupmanagementRejectLeaveRequest(groupRequestId: $groupRequestId) {
                    ok
                    error
                }
            }
            ''',
            variables={
                "groupRequestId": self.request.id,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                "data": {
                    "groupmanagementRejectLeaveRequest": {
                        "ok": True,
                        "error": None,
                    }
                }
            }
        )

        self.assertIn(self.group, self.user2.groups.all())
        self.assertEqual(RequestLog.objects.count(), 1)
        self.assertEqual(GroupRequest.objects.count(), 0)

    def test_reject_request_not_exists(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            mutation($groupRequestId: Int!) {
                groupmanagementRejectLeaveRequest(groupRequestId: $groupRequestId) {
                    ok
                    error
                }
            }
            ''',
            variables={
                "groupRequestId": generate_invalid_pk(GroupRequest),
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                "data": {
                    "groupmanagementRejectLeaveRequest": {
                        "ok": False,
                        "error": "Group request doesn't exist",
                    }
                }
            }
        )

    def test_reject_cannot_manage(self):
        self.client.force_login(self.user2)

        group2 = Group.objects.create(name="Test Group 2")
        group2.authgroup.group_leaders.add(self.user2)

        response = self.query(
            '''
            mutation($groupRequestId: Int!) {
                groupmanagementRejectLeaveRequest(groupRequestId: $groupRequestId) {
                    ok
                    error
                }
            }
            ''',
            variables={
                "groupRequestId": self.request.id,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                "data": {
                    "groupmanagementRejectLeaveRequest": {
                        "ok": False,
                        "error": 'Permission denied',
                    }
                }
            }
        )

        self.assertIn(self.group, self.user2.groups.all())
        self.assertEqual(RequestLog.objects.count(), 0)
        self.assertEqual(GroupRequest.objects.count(), 1)

    @patch.object(GroupRequest, 'delete')
    def test_reject_random_error(self, mock_delete):
        self.client.force_login(self.user)

        mock_delete.side_effect = Exception("Test")

        response = self.query(
            '''
            mutation($groupRequestId: Int!) {
                groupmanagementRejectLeaveRequest(groupRequestId: $groupRequestId) {
                    ok
                    error
                }
            }
            ''',
            variables={
                "groupRequestId": self.request.id,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                "data": {
                    "groupmanagementRejectLeaveRequest": {
                        "ok": False,
                        "error": f'An unhandled error occurred while processing the application from {self.request.main_char} to {self.request.group}.',
                    }
                }
            }
        )

        # Same as test_reject_random_error of other test class

        # self.assertIn(self.group, self.user2.groups.all())
        # self.assertEqual(RequestLog.objects.count(), 0)
        # self.assertEqual(GroupRequest.objects.count(), 1)



import datetime
from graphene_django.utils.testing import GraphQLTestCase

from django.utils import timezone

from allianceauth.tests.auth_utils import AuthUtils
from app_utils.testdata_factories import UserMainFactory, EveCorporationInfoFactory

from allianceauth.timerboard.models import Timer, TimerType

from ..timerboard.types import TimerStructureChoices, TimerTypeChoices, TimerObjectiveChoices


class TestQueries(GraphQLTestCase):
    maxDiff = None

    @classmethod
    def setUpTestData(cls):
        cls.user = AuthUtils.add_permission_to_user_by_name('auth.timer_view', UserMainFactory(), False)

        cls.corp = cls.user.profile.main_character.corporation
        cls.corp2 = EveCorporationInfoFactory()

        cls.corp_future_timer = Timer.objects.create(
            timer_type=TimerType.UNSPECIFIED,
            eve_time=timezone.now() + datetime.timedelta(days=1),
            eve_corp=cls.corp,
            corp_timer=True,
        )

        cls.corp_past_timer = Timer.objects.create(
            timer_type=TimerType.UNSPECIFIED,
            eve_time=timezone.now() - datetime.timedelta(days=1),
            eve_corp=cls.corp,
            corp_timer=True,
        )

        cls.corp2_future_timer = Timer.objects.create(
            timer_type=TimerType.UNSPECIFIED,
            eve_time=timezone.now() + datetime.timedelta(days=1),
            eve_corp=cls.corp2,
            corp_timer=True,
        )

        cls.corp2_past_timer = Timer.objects.create(
            timer_type=TimerType.UNSPECIFIED,
            eve_time=timezone.now() - datetime.timedelta(days=1),
            eve_corp=cls.corp2,
            corp_timer=True,
        )

        cls.future_timer = Timer.objects.create(
            timer_type=TimerType.UNSPECIFIED,
            eve_time=timezone.now() + datetime.timedelta(days=1),
            eve_corp=cls.corp,
            corp_timer=False,
        )

        cls.past_timer = Timer.objects.create(
            timer_type=TimerType.UNSPECIFIED,
            eve_time=timezone.now() - datetime.timedelta(days=1),
            eve_corp=cls.corp,
            corp_timer=False,
        )

    def test_tmr_future_timers(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            query {
                tmrFutureTimers {
                    id
                }
            }
            '''
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'tmrFutureTimers': [
                        {
                            'id': str(self.corp_future_timer.pk),
                        },
                        {
                            'id': str(self.future_timer.pk),
                        },
                    ]
                }
            }
        )

    def test_tmr_past_timers(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            query {
                tmrPastTimers {
                    id
                }
            }
            '''
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'tmrPastTimers': [
                        {
                            'id': str(self.corp_past_timer.pk),
                        },
                        {
                            'id': str(self.past_timer.pk),
                        },
                    ]
                }
            }
        )


class TestMutations(GraphQLTestCase):
    maxDiff = None

    @classmethod
    def setUpTestData(cls):
        cls.user = AuthUtils.add_permission_to_user_by_name('auth.timer_management', UserMainFactory(), False)

        cls.timer = Timer.objects.create(
            timer_type=TimerType.UNSPECIFIED,
            eve_time=timezone.now() + datetime.timedelta(days=1),
            eve_corp=cls.user.profile.main_character.corporation,
            corp_timer=True,
        )

        cls.input_data = {
            'details': 'test',
            'system': 'Jita',
            'structure': TimerStructureChoices.Keepstar.name,
            'timerType': TimerTypeChoices.FINAL.name,
            'objective': TimerObjectiveChoices.Neutral.name,
            'daysLeft': 2,
            'hoursLeft': 0,
            'minutesLeft': 0,
            'important': True,
            'corpTimer': False,
        }

    def test_add_ok(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            mutation($input: TimerInput!) {
                tmrAddTimer(input: $input) {
                    ok
                    timer {
                        details
                        system
                        structure
                        timerType
                        objective
                        important
                        corpTimer
                    }
                }
            }
            ''',
            input_data=self.input_data
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'tmrAddTimer': {
                        'ok': True,
                        'timer': {
                            'details': 'test',
                            'system': 'Jita',
                            'structure': TimerStructureChoices.Keepstar.name,
                            'timerType': TimerTypeChoices.FINAL.name,
                            'objective': TimerObjectiveChoices.Neutral.name,
                            'important': True,
                            'corpTimer': False,
                        }
                    }
                }
            }
        )

    def test_add_error_days(self):
        self.client.force_login(self.user)

        self.input_data['daysLeft'] = -1

        response = self.query(
            '''
            mutation($input: TimerInput!) {
                tmrAddTimer(input: $input) {
                    ok
                    timer {
                        details
                        system
                        structure
                        timerType
                        objective
                        important
                        corpTimer
                    }
                }
            }
            ''',
            input_data=self.input_data
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'tmrAddTimer': {
                        'ok': False,
                        'timer': None,
                    }
                }
            }
        )

        self.assertEqual(Timer.objects.count(), 1)

    def test_add_error_hours(self):
        self.client.force_login(self.user)

        self.input_data['hoursLeft'] = -1

        response = self.query(
            '''
            mutation($input: TimerInput!) {
                tmrAddTimer(input: $input) {
                    ok
                    timer {
                        details
                        system
                        structure
                        timerType
                        objective
                        important
                        corpTimer
                    }
                }
            }
            ''',
            input_data=self.input_data
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'tmrAddTimer': {
                        'ok': False,
                        'timer': None,
                    }
                }
            }
        )

        self.assertEqual(Timer.objects.count(), 1)

    def test_add_error_minutes(self):
        self.client.force_login(self.user)

        self.input_data['minutesLeft'] = -1

        response = self.query(
            '''
            mutation($input: TimerInput!) {
                tmrAddTimer(input: $input) {
                    ok
                    timer {
                        details
                        system
                        structure
                        timerType
                        objective
                        important
                        corpTimer
                    }
                }
            }
            ''',
            input_data=self.input_data
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'tmrAddTimer': {
                        'ok': False,
                        'timer': None,
                    }
                }
            }
        )

        self.assertEqual(Timer.objects.count(), 1)

    def test_edit_ok(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            mutation($timerId: ID!, $input: TimerInput!) {
                tmrEditTimer(timerId: $timerId, input: $input) {
                    ok
                    timer {
                        details
                        system
                        structure
                        timerType
                        objective
                        important
                        corpTimer
                    }
                }
            }
            ''',
            input_data=self.input_data,
            variables={'timerId': self.timer.pk}
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'tmrEditTimer': {
                        'ok': True,
                        'timer': {
                            'details': 'test',
                            'system': 'Jita',
                            'structure': TimerStructureChoices.Keepstar.name,
                            'timerType': TimerTypeChoices.FINAL.name,
                            'objective': TimerObjectiveChoices.Neutral.name,
                            'important': True,
                            'corpTimer': False,
                        }
                    }
                }
            }
        )

    def test_edit_error_days(self):
        self.client.force_login(self.user)

        self.input_data['daysLeft'] = -1

        response = self.query(
            '''
            mutation($timerId: ID!, $input: TimerInput!) {
                tmrEditTimer(timerId: $timerId, input: $input) {
                    ok
                    timer {
                        details
                        system
                        structure
                        timerType
                        objective
                        important
                        corpTimer
                    }
                }
            }
            ''',
            input_data=self.input_data,
            variables={'timerId': self.timer.pk}
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'tmrEditTimer': {
                        'ok': False,
                        'timer': None,
                    }
                }
            }
        )

    def test_edit_error_hours(self):
        self.client.force_login(self.user)

        self.input_data['hoursLeft'] = -1

        response = self.query(
            '''
            mutation($timerId: ID!, $input: TimerInput!) {
                tmrEditTimer(timerId: $timerId, input: $input) {
                    ok
                    timer {
                        details
                        system
                        structure
                        timerType
                        objective
                        important
                        corpTimer
                    }
                }
            }
            ''',
            input_data=self.input_data,
            variables={'timerId': self.timer.pk}
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'tmrEditTimer': {
                        'ok': False,
                        'timer': None,
                    }
                }
            }
        )

    def test_edit_error_minutes(self):
        self.client.force_login(self.user)

        self.input_data['minutesLeft'] = -1

        response = self.query(
            '''
            mutation($timerId: ID!, $input: TimerInput!) {
                tmrEditTimer(timerId: $timerId, input: $input) {
                    ok
                    timer {
                        details
                        system
                        structure
                        timerType
                        objective
                        important
                        corpTimer
                    }
                }
            }
            ''',
            input_data=self.input_data,
            variables={'timerId': self.timer.pk}
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'tmrEditTimer': {
                        'ok': False,
                        'timer': None,
                    }
                }
            }
        )

    def test_delete(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            mutation($timerId: ID!) {
                tmrDeleteTimer(timerId: $timerId) {
                    ok
                }
            }
            ''',
            variables={'timerId': self.timer.pk}
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'tmrDeleteTimer': {
                        'ok': True,
                    }
                }
            }
        )

        self.assertEqual(Timer.objects.count(), 0)



import graphene
from graphql_jwt.decorators import user_passes_test, login_required

from allianceauth.corputils.views import access_corpstats_test
from allianceauth.corputils.models import CorpStats, CorpMember

from .types import CorpStatsType, CorpMemberType


class Query:
    corputils_get_all_corpstats = graphene.List(CorpStatsType)
    corputils_get_corpstats_corp = graphene.Field(CorpStatsType, corp_id=graphene.Int(required=True))
    corputils_search_corpstats = graphene.List(CorpMemberType, search_string=graphene.String(required=True))

    @login_required
    @user_passes_test(access_corpstats_test)
    def resolve_corputils_get_all_corpstats(self, info):
        return CorpStats.objects.visible_to(info.context.user).order_by('corp__corporation_name')

    @login_required
    @user_passes_test(access_corpstats_test)
    def resolve_corputils_get_corpstats_corp(self, info, corp_id):
        avaiable = CorpStats.objects.visible_to(info.context.user)
        stats = CorpStats.objects.get(corp__corporation_id=corp_id)
        if avaiable.filter(pk=stats.pk).exists():
            return stats

    @login_required
    @user_passes_test(access_corpstats_test)
    def resolve_corputils_search_corpstats(self, info, search_string):
        avaiable = CorpStats.objects.visible_to(info.context.user)
        return (
            CorpMember.objects
            .filter(
                corpstats__in=avaiable,
                character_name__icontains=search_string
            )
            .order_by('character_name')
        )



import graphene
from graphql_jwt.decorators import login_required

from allianceauth.notifications.models import Notification

from .types import NotificationType


class Query:
    notif_read_list = graphene.List(NotificationType)
    notif_unread_list = graphene.List(NotificationType)
    notif_unread_count = graphene.Int(user_pk=graphene.ID())

    @login_required
    def resolve_notif_read_list(self, info):
        return Notification.objects.filter(user=info.context.user, viewed=True).order_by("-timestamp")

    @login_required
    def resolve_notif_unread_list(self, info):
        return Notification.objects.filter(user=info.context.user, viewed=False).order_by("-timestamp")

    def resolve_notif_unread_count(self, info, user_pk=None):
        if user_pk is None and not info.context.user.is_authenticated:
            return -1

        pk = user_pk if user_pk is not None else info.context.user.pk
        return Notification.objects.user_unread_count(pk)



from graphene_django.utils.testing import GraphQLTestCase

from allianceauth.notifications import notify
from allianceauth.notifications.models import Notification

from app_utils.testdata_factories import UserFactory


class TestQueries(GraphQLTestCase):
    maxDiff = None

    @classmethod
    def setUpTestData(cls):
        cls.user = UserFactory()

        notify(cls.user, "Test notif 1", level="success")

        cls.notif1: Notification = Notification.objects.get(title="Test notif 1")

        notify(cls.user, "Test notif 2", level="warning")

        cls.notif2: Notification = Notification.objects.get(title="Test notif 2")

        cls.notif1.viewed = True
        cls.notif1.save()

    def test_notif_read_list(self):
        self.client.force_login(self.user, "graphql_jwt.backends.JSONWebTokenBackend")

        response = self.query(
            '''
            query q {
                notifReadList {
                    id
                }
            }
            '''
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'notifReadList': [
                        {
                            'id': str(self.notif1.pk)
                        }
                    ]
                }
            }
        )

    def test_notif_unread_list(self):
        self.client.force_login(self.user, "graphql_jwt.backends.JSONWebTokenBackend")

        response = self.query(
            '''
            query q {
                notifUnreadList {
                    id
                }
            }
            '''
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'notifUnreadList': [
                        {
                            'id': str(self.notif2.pk)
                        }
                    ]
                }
            }
        )

    def test_notif_unread_count_user_pk(self):
        response = self.query(
            '''
            query q($input: ID!) {
                notifUnreadCount(userPk: $input)
            }
            ''',
            input_data=self.user.pk
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'notifUnreadCount': 1
                }
            }
        )

    def test_notif_unread_count_login(self):
        self.client.force_login(self.user, "graphql_jwt.backends.JSONWebTokenBackend")

        response = self.query(
            '''
            query q {
                notifUnreadCount
            }
            '''
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'notifUnreadCount': 1
                }
            }
        )

    def test_notif_unread_count_error(self):
        response = self.query(
            '''
            query q {
                notifUnreadCount
            }
            '''
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'notifUnreadCount': -1
                }
            }
        )


class TestMutations(GraphQLTestCase):
    maxDiff = None

    @classmethod
    def setUpTestData(cls):
        cls.user = UserFactory()
        cls.user2 = UserFactory()

        notify(cls.user, "Test notif 1", level="success")

        cls.notif1: Notification = Notification.objects.get(title="Test notif 1")

        notify(cls.user, "Test notif 2", level="warning")

        cls.notif2: Notification = Notification.objects.get(title="Test notif 2")

    def test_mark_read_mutation_ok(self):
        self.client.force_login(self.user, "graphql_jwt.backends.JSONWebTokenBackend")

        response = self.query(
            '''
            mutation m($input: ID!) {
                notifMarkAsRead(notifId: $input) {
                    notification {
                        id
                        viewed
                    }
                }
            }
            ''',
            input_data=self.notif1.pk
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'notifMarkAsRead': {
                        'notification': {
                            'id': str(self.notif1.pk),
                            'viewed': True
                        }
                    }
                }
            }
        )

    def test_mark_read_mutation_not_ok(self):
        self.client.force_login(self.user2, "graphql_jwt.backends.JSONWebTokenBackend")

        response = self.query(
            '''
            mutation m($input: ID!) {
                notifMarkAsRead(notifId: $input) {
                    ok
                    notification {
                        id
                        viewed
                    }
                }
            }
            ''',
            input_data=self.notif1.pk
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'notifMarkAsRead': {
                        'notification': None,
                        'ok': False
                    }
                }
            }
        )

    def test_delete_mutation_ok(self):
        self.client.force_login(self.user, "graphql_jwt.backends.JSONWebTokenBackend")

        response = self.query(
            '''
            mutation m($input: ID!) {
                notifDelete(notifId: $input) {
                    ok
                }
            }
            ''',
            input_data=self.notif1.pk
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'notifDelete': {
                        'ok': True
                    }
                }
            }
        )

        self.assertFalse(
            Notification.objects
            .filter(pk=self.notif1.pk)
            .exists()
        )

    def test_delete_mutation_not_ok(self):
        self.client.force_login(self.user2, "graphql_jwt.backends.JSONWebTokenBackend")

        response = self.query(
            '''
            mutation m($input: ID!) {
                notifDelete(notifId: $input) {
                    ok
                }
            }
            ''',
            input_data=self.notif1.pk
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'notifDelete': {
                        'ok': False
                    }
                }
            }
        )

        self.assertTrue(
            Notification.objects
            .filter(pk=self.notif1.pk)
            .exists()
        )

    def test_all_read(self):
        self.client.force_login(self.user, "graphql_jwt.backends.JSONWebTokenBackend")

        response = self.query(
            '''
            mutation m {
                notifMarkAllRead {
                    ok
                }
            }
            '''
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'notifMarkAllRead': {
                        'ok': True
                    }
                }
            }
        )

        self.assertFalse(
            Notification.objects
            .filter(user=self.user, viewed=False)
            .exists()
        )

        self.assertEqual(
            Notification.objects
            .filter(user=self.user, viewed=True)
            .count(),
            2
        )

    def test_delete_all_read(self):
        self.client.force_login(self.user, "graphql_jwt.backends.JSONWebTokenBackend")

        self.notif1.viewed = True
        self.notif1.save()

        response = self.query(
            '''
            mutation m {
                notifDeleteAllRead {
                    ok
                }
            }
            '''
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'notifDeleteAllRead': {
                        'ok': True
                    }
                }
            }
        )

        self.assertFalse(
            Notification.objects
            .filter(pk=self.notif1.pk)
            .exists()
        )

        self.assertTrue(
            Notification.objects
            .filter(pk=self.notif2.pk)
            .exists()
        )



from graphene_django import DjangoObjectType

from allianceauth.optimer.models import OpTimer, OpTimerType


class OpTimerModelType(DjangoObjectType):
    class Meta:
        model = OpTimer


class OpTimerTypeType(DjangoObjectType):
    class Meta:
        model = OpTimerType
        fields = ('id', "type", )



import datetime

import graphene
from graphql_jwt.decorators import login_required, permission_required

from django.contrib.auth import get_user_model
from django.db.models.functions import ExtractMonth, ExtractYear
from django.db.models import Count

from allianceauth.fleetactivitytracking.models import Fatlink, Fat
from allianceauth.fleetactivitytracking.views import MemberStat, first_day_of_next_month, CorpStat
from allianceauth.eveonline.models import EveCorporationInfo, EveCharacter

from .types import FatlinkType, FatType, FatUserStatsType, FatCorpStatsType, FatPersonalStatsType, FatPersonalMonthlyStatsType

User = get_user_model()


class Query:
    fat_recent_fat = graphene.List(FatType, num=graphene.Int())
    fat_get_fatlinks = graphene.List(FatlinkType, num=graphene.Int())
    fat_corp_monthly_stats = graphene.List(FatUserStatsType, corp_id=graphene.Int(required=True), year=graphene.Int(required=True), month=graphene.Int(required=True))
    fat_general_monthly_stats = graphene.List(FatCorpStatsType, year=graphene.Int(required=True), month=graphene.Int(required=True))
    fat_personal_stats = graphene.List(FatPersonalStatsType)
    fat_personal_monthly_stats = graphene.Field(FatPersonalMonthlyStatsType, year=graphene.Int(required=True), month=graphene.Int(required=True), char_id=graphene.Int())

    @login_required
    def resolve_fat_recent_fat(self, info, num=5):
        return Fat.objects.filter(user=info.context.user).order_by('-id')[:num]

    @login_required
    @permission_required('auth.fleetactivitytracking')
    def resolve_fat_get_fatlinks(self, info, num=5):
        return Fatlink.objects.order_by('-id')[:num]

    @login_required
    @permission_required('auth.fleetactivitytracking_statistics')
    def resolve_fat_corp_monthly_stats(self, info, corp_id, year, month):
        start_of_month = datetime.datetime(year, month, 1)
        start_of_next_month = first_day_of_next_month(year, month)

        users = User.objects.select_related('profile__main_character').filter(character_ownerships__character__corporation_id=corp_id).distinct()

        stat_list = []

        for member in users:
            stat = MemberStat(member, start_of_month, start_of_next_month)
            stat_list.append(
                {
                    'user': member,
                    'num_chars': stat.n_chars,
                    'num_fats': stat.n_fats,
                    'avg_fats': stat.avg_fat
                }
            )

        stat_list.sort(key=lambda stat: stat['user'].profile.main_character.character_name)
        stat_list.sort(key=lambda stat: (stat['num_fats'], stat['avg_fats']), reverse=True)

        return stat_list

    @login_required
    @permission_required('auth.fleetactivitytracking_statistics')
    def resolve_fat_general_monthly_stats(self, info, year, month):
        start_of_month = datetime.datetime(year, month, 1)
        start_of_next_month = first_day_of_next_month(year, month)

        stat_list = []

        for corp in EveCorporationInfo.objects.all():
            stat = CorpStat(corp.corporation_id, start_of_month, start_of_next_month)
            stat_list.append(
                {
                    'corporation': corp,
                    'num_fats': stat.n_fats,
                    'avg_fats': stat.avg_fat
                }
            )

        # sort stats
        stat_list.sort(key=lambda stat: stat['corporation'].corporation_name)
        stat_list.sort(key=lambda stat: (stat['num_fats'], stat['avg_fats']), reverse=True)

        return stat_list

    @login_required
    def resolve_fat_personal_stats(self, info):
        # my optimized code
        return (
            Fat.objects.filter(user=info.context.user)
            .annotate(month=ExtractMonth('fatlink__fatdatetime'))
            .annotate(year=ExtractYear('fatlink__fatdatetime'))
            .values('month', 'year')
            .annotate(num_fats=Count('*'))
            .order_by('-year', '-month')
        )

    @login_required
    def resolve_fat_personal_monthly_stats(self, info, year, month, char_id=None):
        if char_id and info.context.user.has_perm('auth.fleetactivitytracking_statistics'):
            user = EveCharacter.objects.get(character_id=char_id).character_ownership.user
        else:
            user = info.context.user

        # my optimized code
        personal_fats = Fat.objects.filter(user=user, fatlink__fatdatetime__year=year, fatlink__fatdatetime__month=month)\
            .values('shiptype')\
            .annotate(times_used=Count('*'))\
            .order_by('shiptype')

        created_fats = Fatlink.objects.filter(creator=user, fatdatetime__year=year, fatdatetime__month=month)

        return {
            'collected_links': personal_fats,
            'created_links': created_fats,
        }



import graphene

from allianceauth.authentication.models import CharacterOwnership


class EntryRoleInput(graphene.InputObjectType):
    name = graphene.String(required=True)
    value = graphene.Int(required=True)


class EntryCharacterInput(graphene.InputObjectType):
    site_count = graphene.Int(required=True)
    user_id = graphene.Int(required=True)
    character_id = graphene.Int(required=True)
    helped_setup = graphene.Boolean(default_value=False)
    role = graphene.String(required=True)


class EntryInput(graphene.InputObjectType):
    estimated_total = graphene.Int(required=True)
    shares = graphene.List(EntryCharacterInput, required=True)
    roles = graphene.List(EntryRoleInput, required=True)

    @staticmethod
    def is_valid(input):
        errors = []

        total = 0
        roles = {}
        for new_role in input.roles:
            if new_role['name'] in roles:
                errors.append(f"{new_role['name']} name is not unique")
            else:
                roles[new_role['name']] = new_role['value']

        characters_ids = set()
        for new_share in input.shares:
            if new_share['character_id'] in characters_ids:
                errors.append(f"character {new_share['character_id']} cannot have more than 1 share")
            else:
                characters_ids.add(new_share['character_id'])

            if new_share['role'] in roles:
                total += roles[new_share['role']] * new_share['site_count']
            else:
                errors.append(f"{new_share['role']} is not a valid role")

            if not CharacterOwnership.objects.filter(user_id=new_share['user_id'], character_id=new_share['character_id']).exists():
                errors.append("character ownership doesn't match")

        if total <= 0:
            errors.append('Form not valid, you need at least 1 person to receive loot')

        if len(input.roles) == 0 or len(input.shares) == 0:
            errors.append('Not enough shares or roles')

        return errors


class RotationCloseInput(graphene.InputObjectType):
    rotation_id = graphene.Int(required=True)
    sales_value = graphene.Int(required=True)


class CharactersStatsInput(graphene.InputObjectType):
    start_date = graphene.Date(required=True)
    end_date = graphene.Date(required=True)



import datetime

import graphene
from graphql_jwt.decorators import login_required, permission_required
from graphene_django.forms.mutation import DjangoFormMutation

from django.utils import timezone
from django.core.exceptions import ValidationError
from django.utils.crypto import get_random_string

from allianceauth.fleetactivitytracking.models import Fatlink, Fat
from allianceauth.fleetactivitytracking.views import SWAGGER_SPEC_PATH
from allianceauth.fleetactivitytracking.forms import FatlinkForm
from allianceauth.eveonline.models import EveCharacter
from allianceauth.eveonline.providers import provider
from esi.models import Token

from .types import FatlinkType
from ..decorators import tokens_required


class AddFatParticipation(graphene.Mutation):
    _required_scopes = [
        'esi-location.read_location.v1',
        'esi-location.read_ship_type.v1',
        'esi-universe.read_structures.v1',
        'esi-location.read_online.v1',
    ]

    class Arguments:
        token_id = graphene.ID(required=True)
        fatlink_hash = graphene.String(required=True)

    ok = graphene.Boolean()
    error = graphene.String()

    @classmethod
    @login_required
    @tokens_required(scopes=_required_scopes)
    def mutate(cls, root, info, token_id, fatlink_hash):
        try:
            token = Token.objects.all().require_scopes(' '.join(cls._required_scopes)).get(pk=token_id)
            c = token.get_esi_client(spec_file=SWAGGER_SPEC_PATH)
            character_online = c.Location.get_characters_character_id_online(
                character_id=token.character_id
            ).result()
            character = EveCharacter.objects.get_character_by_id(token.character_id)
            error = None
        except Token.DoesNotExist:
            error = 'Token not valid'
            ok = False

        if not error and character_online["online"] is True:
            fatlink = Fatlink.objects.get(hash=fatlink_hash)
            if (token.user == info.context.user) and (timezone.now() - fatlink.fatdatetime) < datetime.timedelta(seconds=(fatlink.duration * 60)):
                if character:
                    location = c.Location.get_characters_character_id_location(character_id=token.character_id).result()
                    ship = c.Location.get_characters_character_id_ship(character_id=token.character_id).result()
                    location['solar_system_name'] = \
                        c.Universe.get_universe_systems_system_id(system_id=location['solar_system_id']).result()['name']
                    if location['station_id']:
                        location['station_name'] = \
                            c.Universe.get_universe_stations_station_id(station_id=location['station_id']).result()['name']
                    elif location['structure_id']:
                        location['station_name'] = \
                            c.Universe.get_universe_structures_structure_id(structure_id=location['structure_id']).result()[
                                'name']
                    else:
                        location['station_name'] = "No Station"
                    ship['ship_type_name'] = provider.get_itemtype(ship['ship_type_id']).name

                    fat = Fat()
                    fat.system = location['solar_system_name']
                    fat.station = location['station_name']
                    fat.shiptype = ship['ship_type_name']
                    fat.fatlink = fatlink
                    fat.character = character
                    fat.user = info.context.user
                    try:
                        fat.full_clean()
                        fat.save()
                        ok = True
                        error = None
                    except ValidationError as e:
                        err_messages = []
                        for errorname, message in e.message_dict.items():
                            err_messages.append(message[0])
                        error = ' '.join(err_messages)
                        ok = False
                else:
                    ok = False
                    error = "Character doesn't exists"
            else:
                ok = False
                error = "FAT link has expired or user not valid"

        elif not error:
            ok = False
            error = f"Cannot register the fleet participation for {character.character_name}. The character needs to be online."

        return cls(ok=ok, error=error)


class CreateFatlink(DjangoFormMutation):
    class Meta:
        form_class = FatlinkForm

    ok = graphene.Boolean()
    fatlink = graphene.Field(FatlinkType)

    @classmethod
    @login_required
    @permission_required('auth.fleetactivitytracking')
    def perform_mutate(cls, form: FatlinkForm, info):
        fatlink = Fatlink()
        fatlink.fleet = form.cleaned_data["fleet"]
        fatlink.duration = form.cleaned_data["duration"]
        fatlink.fatdatetime = timezone.now()
        fatlink.creator = info.context.user
        fatlink.hash = get_random_string(length=15)
        fatlink.save()

        return cls(ok=True, fatlink=fatlink, **form.cleaned_data)


class RemoveCharFatlink(graphene.Mutation):
    class Arguments:
        fatlink_hash = graphene.String(required=True)
        character_id = graphene.Int(required=True)

    ok = graphene.Boolean()
    fatlink = graphene.Field(FatlinkType)

    @classmethod
    @login_required
    @permission_required('auth.fleetactivitytracking')
    def mutate(cls, root, info, fatlink_hash, character_id):
        fatlink = Fatlink.objects.get(hash=fatlink_hash)
        Fat.objects.filter(fatlink=fatlink, character__character_id=character_id).delete()
        return cls(ok=True, fatlink=fatlink)


class DeleteFatlink(graphene.Mutation):
    class Arguments:
        fatlink_hash = graphene.String(required=True)

    ok = graphene.Boolean()

    @classmethod
    @login_required
    @permission_required('auth.fleetactivitytracking')
    def mutate(cls, root, info, fatlink_hash):
        Fatlink.objects.get(hash=fatlink_hash).delete()
        return cls(ok=True)


class Mutation:
    fat_participate_to_fatlink = AddFatParticipation.Field()
    fat_create_fatlink = CreateFatlink.Field()
    fat_remove_char_fat = RemoveCharFatlink.Field()
    fat_delete_fatlink = DeleteFatlink.Field()



import json
from graphene_django.utils.testing import GraphQLTestCase
from unittest.mock import patch

from django.db.models import Max

from allianceauth.tests.test_auth_utils import AuthUtils
from app_utils.testdata_factories import UserMainFactory, EveCharacterFactory
from app_utils.esi_testing import EsiClientStub, EsiEndpoint

from allianceauth.corputils.models import CorpStats, CorpMember


class TestQueriesAndTypes(GraphQLTestCase):
    maxDiff = None

    @classmethod
    def setUpTestData(cls):
        cls.user, cls.user2 = UserMainFactory.create_batch(2)

        cls.user = AuthUtils.add_permissions_to_user_by_name(
            [
                'corputils.add_corpstats',
                'corputils.view_corp_corpstats',
            ],
            cls.user,
            False
        )

        cls.mainchar = cls.user.profile.main_character
        cls.corp = cls.mainchar.corporation

        cls.newchars = EveCharacterFactory.create_batch(9, corporation=cls.corp)

        cls.corpstat = CorpStats.objects.create(
            token=cls.user.token_set.first(),
            corp=cls.corp
        )

        cls.corpstat2 = CorpStats.objects.create(
            token=cls.user2.token_set.first(),
            corp=cls.user2.profile.main_character.corporation
        )

        CorpMember.objects.bulk_create(
            [
                CorpMember(
                    character_id=char.character_id,
                    character_name=char.character_name,
                    corpstats=cls.corpstat
                ) for char in cls.newchars
            ]
        )

        CorpMember.objects.create(
            character_id=cls.mainchar.character_id,
            character_name=cls.mainchar.character_name,
            corpstats=cls.corpstat
        )

    def test_get_corpstats(self):
        self.client.force_login(self.user, "graphql_jwt.backends.JSONWebTokenBackend")

        response = self.query(
            '''
            query q {
                corputilsGetAllCorpstats {
                    corp {
                        id
                    }
                }
            }
            '''
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'corputilsGetAllCorpstats': [
                        {
                            'corp': {
                                'id': str(self.corp.pk)
                            }
                        }
                    ]
                }
            }
        )

    def test_get_corpstats_corp_ok(self):
        self.client.force_login(self.user, "graphql_jwt.backends.JSONWebTokenBackend")

        response = self.query(
            '''
            query q($input: Int!) {
                corputilsGetCorpstatsCorp(corpId: $input) {
                    corp {
                        id
                    }
                }
            }
            ''',
            input_data=self.corp.corporation_id
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'corputilsGetCorpstatsCorp': {
                        'corp': {
                            'id': str(self.corp.pk)
                        }
                    }
                }
            }
        )

    def test_get_corpstats_corp_not_ok(self):
        self.client.force_login(self.user, "graphql_jwt.backends.JSONWebTokenBackend")

        response = self.query(
            '''
            query q($input: Int!) {
                corputilsGetCorpstatsCorp(corpId: $input) {
                    corp {
                        id
                    }
                }
            }
            ''',
            input_data=self.corpstat2.corp.corporation_id
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'corputilsGetCorpstatsCorp': None
                }
            }
        )

    def test_search_corpstats(self):
        self.client.force_login(self.user, "graphql_jwt.backends.JSONWebTokenBackend")

        response = self.query(
            '''
            query q($input: String!) {
                corputilsSearchCorpstats(searchString: $input) {
                    characterId
                }
            }
            ''',
            input_data=self.mainchar.character_name
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'corputilsSearchCorpstats': [
                        {
                            'characterId': self.mainchar.character_id
                        }
                    ]
                }
            }
        )

    def test_character_field_ok(self):
        self.client.force_login(self.user, "graphql_jwt.backends.JSONWebTokenBackend")

        response = self.query(
            '''
            query q($input: String!) {
                corputilsSearchCorpstats(searchString: $input) {
                    characterId
                    character {
                        id
                    }
                }
            }
            ''',
            input_data=self.mainchar.character_name
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'corputilsSearchCorpstats': [
                        {
                            'characterId': self.mainchar.character_id,
                            'character': {
                                'id': str(self.mainchar.pk)
                            }
                        }
                    ]
                }
            }
        )

    def test_character_field_not_ok(self):
        self.client.force_login(self.user, "graphql_jwt.backends.JSONWebTokenBackend")

        max_char_id = CorpMember.objects.aggregate(m=Max('character_id'))['m']

        notregisteredmember = CorpMember.objects.create(
            character_id=max_char_id + 1,
            character_name=self.mainchar.character_name + ' 2',
            corpstats=self.corpstat
        )

        response = self.query(
            '''
            query q($input: String!) {
                corputilsSearchCorpstats(searchString: $input) {
                    characterId
                    character {
                        id
                    }
                }
            }
            ''',
            input_data=notregisteredmember.character_name
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'corputilsSearchCorpstats': [
                        {
                            'characterId': max_char_id + 1,
                            'character': None
                        }
                    ]
                }
            }
        )

    def test_registered_type(self):
        self.client.force_login(self.user, "graphql_jwt.backends.JSONWebTokenBackend")

        response = self.query(
            '''
            query q {
                corputilsGetAllCorpstats {
                    registered {
                        id
                    }
                }
            }
            '''
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'corputilsGetAllCorpstats': [
                        {
                            'registered': [
                                {
                                    'id': str(self.mainchar.pk)
                                }
                            ]
                        }
                    ]
                }
            }
        )

    def test_unregistered_type(self):
        self.client.force_login(self.user, "graphql_jwt.backends.JSONWebTokenBackend")

        response = self.query(
            '''
            query q {
                corputilsGetAllCorpstats {
                    unregistered {
                        id
                    }
                }
            }
            '''
        )

        expected = [str(char.pk) for char in self.newchars]

        data = json.loads(response.content)

        self.assertIn('data', data)
        self.assertNotIn('errors', data)
        self.assertIn('corputilsGetAllCorpstats', data['data'])

        results = data['data']['corputilsGetAllCorpstats']

        self.assertEqual(len(results), 1)

        results = results[0]['unregistered']

        self.assertCountEqual(
            [r['id'] for r in results],
            expected
        )

    def test_mains_type(self):
        self.client.force_login(self.user, "graphql_jwt.backends.JSONWebTokenBackend")

        response = self.query(
            '''
            query q {
                corputilsGetAllCorpstats {
                    mains {
                        id
                    }
                }
            }
            '''
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'corputilsGetAllCorpstats': [
                        {
                            'mains': [
                                {
                                    'id': str(self.mainchar.pk)
                                }
                            ]
                        }
                    ]
                }
            }
        )


class TestMutations(GraphQLTestCase):
    maxDiff = None

    @classmethod
    def setUpTestData(cls):
        cls.user = UserMainFactory(main_character__scopes=['esi-corporations.read_corporation_membership.v1'])

        cls.user = AuthUtils.add_permission_to_user_by_name(
            'corputils.add_corpstats',
            cls.user,
            False
        )

        cls.token = cls.user.token_set.first()

    @patch('allianceauth.corputils.models.CorpStats.update')
    def test_add_corp_stats_mutation_ok(self, mock_update):
        mock_update.return_value = None

        self.client.force_login(self.user, "graphql_jwt.backends.JSONWebTokenBackend")

        response = self.query(
            '''
            mutation m($input: ID!) {
                corputilsAddCorpstats(tokenId: $input) {
                    ok
                }
            }
            ''',
            input_data=self.token.pk
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'corputilsAddCorpstats': {
                        'ok': True
                    }
                }
            }
        )

        self.assertTrue(mock_update.called)

    def test_add_corp_stats_mutation_wrong_user(self):
        user2 = UserMainFactory()

        self.client.force_login(self.user, "graphql_jwt.backends.JSONWebTokenBackend")

        response = self.query(
            '''
            mutation m($input: ID!) {
                corputilsAddCorpstats(tokenId: $input) {
                    ok
                }
            }
            ''',
            input_data=user2.token_set.first().pk
        )

        data = json.loads(response.content)

        self.assertIn('data', data)
        self.assertIn('errors', data)
        self.assertIn('corputilsAddCorpstats', data['data'])
        self.assertIsNone(data['data']['corputilsAddCorpstats'])

        self.assertEqual(len(data['errors']), 1)

        error = data['errors'][0]

        self.assertIn('locations', error)
        self.assertIn('path', error)
        self.assertIn('message', error)

        self.assertEqual(error['path'], ['corputilsAddCorpstats'])

        self.assertEqual(error['message'], 'Token not valid')

    def test_add_corp_stats_mutation_token_missing(self):
        user2 = UserMainFactory()

        user2 = AuthUtils.add_permission_to_user_by_name(
            'corputils.add_corpstats',
            user2,
            False
        )

        self.client.force_login(user2, "graphql_jwt.backends.JSONWebTokenBackend")

        response = self.query(
            '''
            mutation m($input: ID!) {
                corputilsAddCorpstats(tokenId: $input) {
                    ok
                }
            }
            ''',
            input_data=user2.token_set.first().pk
        )

        data = json.loads(response.content)

        self.assertIn('data', data)
        self.assertIn('errors', data)
        self.assertIn('corputilsAddCorpstats', data['data'])
        self.assertIsNone(data['data']['corputilsAddCorpstats'])

        self.assertEqual(len(data['errors']), 1)

        error = data['errors'][0]

        self.assertIn('locations', error)
        self.assertIn('path', error)
        self.assertIn('message', error)

        self.assertEqual(error['path'], ['corputilsAddCorpstats'])

        self.assertEqual(error['message'], 'Required token missing')

    @patch('esi.models.Token.get_esi_client')
    @patch('allianceauth.corputils.models.CorpStats.update')
    def test_add_corp_stats_mutation_character_not_exists(self, mock_update, mock_esi_client):
        mock_update.return_value = None

        mock_endpoint = EsiEndpoint(
            'Character',
            'get_characters_character_id',
            'character_id',
            data={
                str(self.token.character_id): {
                    'corporation_id': self.user.profile.main_character.corporation_id
                }
            }
        )

        mock_esi_client.return_value = EsiClientStub.create_from_endpoints([mock_endpoint])

        self.client.force_login(self.user, "graphql_jwt.backends.JSONWebTokenBackend")

        self.user.profile.main_character.delete()

        response = self.query(
            '''
            mutation m($input: ID!) {
                corputilsAddCorpstats(tokenId: $input) {
                    ok
                }
            }
            ''',
            input_data=self.token.pk
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'corputilsAddCorpstats': {
                        'ok': True
                    }
                }
            }
        )

        self.assertTrue(mock_update.called)

    @patch('esi.clients.esi_client_factory')
    @patch('allianceauth.corputils.models.CorpStats.update')
    def test_add_corp_stats_mutation_corp_not_exists(self, mock_update, mock_esi_client):
        mock_update.return_value = None

        corp = self.user.profile.main_character.corporation

        mock_endpoint = EsiEndpoint(
            'Corporation',
            'get_corporations_corporation_id',
            'corporation_id',
            data={
                str(corp.corporation_id): {
                    'name': corp.corporation_name,
                    'ticker': corp.corporation_ticker,
                    'member_count': corp.member_count,
                    'alliance_id': corp.alliance_id,
                }
            }
        )

        mock_esi_client.return_value = EsiClientStub.create_from_endpoints([mock_endpoint])

        corp.delete()

        self.client.force_login(self.user, "graphql_jwt.backends.JSONWebTokenBackend")

        response = self.query(
            '''
            mutation m($input: ID!) {
                corputilsAddCorpstats(tokenId: $input) {
                    ok
                }
            }
            ''',
            input_data=self.token.pk
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'corputilsAddCorpstats': {
                        'ok': True
                    }
                }
            }
        )

        self.assertTrue(mock_update.called)

    @patch('allianceauth.corputils.models.CorpStats.update')
    def test_update_corpstats_ok(self, mock_update):
        mock_update.return_value = None

        self.client.force_login(self.user, "graphql_jwt.backends.JSONWebTokenBackend")

        corp = self.user.profile.main_character.corporation

        CorpStats.objects.create(
            token=self.token,
            corp=corp
        )

        response = self.query(
            '''
            mutation m($input: Int!) {
                corputilsUpdateCorpstats(corpId: $input) {
                    ok
                }
            }
            ''',
            input_data=corp.corporation_id
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'corputilsUpdateCorpstats': {
                        'ok': True
                    }
                }
            }
        )

        self.assertTrue(mock_update.called)



import datetime

import graphene
from graphql_jwt.decorators import login_required, permission_required

from django.utils import timezone

from allianceauth.timerboard.models import Timer

from .inputs import TimerInput
from .types import StructureTimerType


class AddTimerMutation(graphene.Mutation):
    class Arguments:
        input = TimerInput(required=True)

    ok = graphene.Boolean()
    timer = graphene.Field(StructureTimerType)

    @classmethod
    @login_required
    @permission_required('auth.timer_management')
    def mutate(cls, root, info, input):
        if input.days_left < 0:
            return cls(ok=False)
        if input.hours_left < 0 or input.hours_left > 23:
            return cls(ok=False)
        if input.minutes_left < 0 or input.minutes_left > 59:
            return cls(ok=False)

        user = info.context.user
        char = user.profile.main_character

        timer = Timer()

        timer.details = input.details
        timer.system = input.system
        timer.planet_moon = input.planet_moon
        timer.structure = input.structure
        timer.timer_type = input.timer_type
        timer.objective = input.objective

        timer.eve_time = timezone.now() + datetime.timedelta(
            days=input.days_left,
            hours=input.hours_left,
            minutes=input.minutes_left
        )

        timer.important = input.important
        timer.corp_timer = input.corp_timer
        timer.user = user
        timer.eve_character = char
        timer.eve_corp = char.corporation

        timer.save()

        return cls(ok=True, timer=timer)


class EditTimerMutation(graphene.Mutation):
    class Arguments:
        input = TimerInput(required=True)
        timer_id = graphene.ID(required=True)

    ok = graphene.Boolean()
    timer = graphene.Field(StructureTimerType)

    @classmethod
    @login_required
    @permission_required('auth.timer_management')
    def mutate(cls, root, info, input, timer_id):
        if input.days_left < 0:
            return cls(ok=False)
        if input.hours_left < 0 or input.hours_left > 23:
            return cls(ok=False)
        if input.minutes_left < 0 or input.minutes_left > 59:
            return cls(ok=False)

        user = info.context.user
        char = user.profile.main_character

        timer = Timer.objects.get(pk=timer_id)

        timer.details = input.details
        timer.system = input.system
        timer.planet_moon = input.planet_moon
        timer.structure = input.structure
        timer.timer_type = input.timer_type
        timer.objective = input.objective

        timer.eve_time = timezone.now() + datetime.timedelta(
            days=input.days_left,
            hours=input.hours_left,
            minutes=input.minutes_left
        )

        timer.important = input.important
        timer.corp_timer = input.corp_timer
        timer.user = user
        timer.eve_character = char
        timer.eve_corp = char.corporation

        timer.save()

        return cls(ok=True, timer=timer)


class DeleteTimerMutation(graphene.Mutation):
    class Arguments:
        timer_id = graphene.ID(required=True)

    ok = graphene.Boolean()

    @classmethod
    @login_required
    @permission_required('auth.timer_management')
    def mutate(cls, root, info, timer_id):
        Timer.objects.filter(pk=timer_id).delete()
        return cls(ok=True)


class Mutation:
    tmr_add_timer = AddTimerMutation.Field()
    tmr_edit_timer = EditTimerMutation.Field()
    tmr_delete_timer = DeleteTimerMutation.Field()



import graphene
from graphql_jwt.decorators import login_required, user_passes_test

from django.contrib.auth.models import Group
from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist, PermissionDenied

from allianceauth.groupmanagement.managers import GroupManager
from allianceauth.groupmanagement.models import RequestLog, GroupRequest
from allianceauth.services.hooks import get_extension_logger
from allianceauth.notifications import notify

from .types import GroupRequestAddStatus, GroupRequestLeaveStatus


logger = get_extension_logger(__name__)


class AddGroupRequest(graphene.Mutation):
    class Arguments:
        group_id = graphene.Int(required=True)

    ok = graphene.Boolean()
    status = graphene.Field(GroupRequestAddStatus)

    @classmethod
    @login_required
    def mutate(cls, root, info, group_id):
        user = info.context.user
        group = Group.objects.get(id=group_id)

        logger.debug(f"group_request_add called by user {user} for group id {group_id}")

        if not GroupManager.joinable_group(group, user.profile.state):
            logger.warning(f"User {user} attempted to join group id {group_id} but it is not a joinable group")
            status = GroupRequestAddStatus.CANNOT_JOIN
            ok = False
        elif group in user.groups.all():
            logger.warning(f"User {user} attempted to join group id {group_id} but they are already a member.")
            status = GroupRequestAddStatus.ALREADY_MEMBER
            ok = False
        elif not user.has_perm('groupmanagement.request_groups') and not group.authgroup.public:
            logger.warning(f"User {user} attempted to join group id {group_id} but it is not a public group")
            status = GroupRequestAddStatus.CANNOT_JOIN
            ok = False
        elif group.authgroup.open:
            logger.info(f"{user} joining {group} as is an open group")
            user.groups.add(group)
            request_info = user.username + ":" + group.name
            log = RequestLog(request_type=False, group=group, request_info=request_info, action=1, request_actor=user)
            log.save()

            status = GroupRequestAddStatus.JOINED
            ok = True
        elif GroupRequest.objects.filter(user=user, group=group).exists():
            logger.info(f"{user} attempted to join {group} but already has an open application")
            status = GroupRequestAddStatus.ALREADY_APPLIED
            ok = False
        else:
            grouprequest = GroupRequest()
            grouprequest.group = group
            grouprequest.user = user
            grouprequest.leave_request = False
            grouprequest.save()
            logger.info(f"Created group request for user {user} to group {group}")
            grouprequest.notify_leaders()

            status = GroupRequestAddStatus.APPLIED
            ok = True

        return cls(ok=ok, status=status)


class LeaveGroupRequest(graphene.Mutation):
    class Arguments:
        group_id = graphene.Int(required=True)

    ok = graphene.Boolean()
    status = graphene.Field(GroupRequestLeaveStatus)

    @classmethod
    @login_required
    def mutate(cls, root, info, group_id):
        user = info.context.user
        logger.debug(f"group_request_leave called by user {user} for group id {group_id}")
        group = Group.objects.get(id=group_id)

        if not GroupManager.check_internal_group(group):
            logger.warning(f"User {user} attempted to leave group id {group_id} but it is not a joinable group")
            status = GroupRequestLeaveStatus.CANNOT_LEAVE
            ok = False
        elif group not in user.groups.all():
            logger.debug(f"User {user} attempted to leave group id {group_id} but they are not a member")
            status = GroupRequestLeaveStatus.NOT_MEMBER
            ok = False
        elif group.authgroup.open:
            logger.info(f"{user} leaving {group} as is an open group")
            request_info = user.username + ":" + group.name
            log = RequestLog(request_type=True, group=group, request_info=request_info, action=1, request_actor=user)
            log.save()
            user.groups.remove(group)

            status = GroupRequestLeaveStatus.LEFT
            ok = True
        elif GroupRequest.objects.filter(user=user, group=group).exists():
            logger.info(f"{user} attempted to leave {group} but already has an pending leave request.")
            status = GroupRequestLeaveStatus.PENDING_LEAVE_REQUEST
            ok = False
        elif getattr(settings, 'GROUPMANAGEMENT_AUTO_LEAVE', False):
            logger.info(f"{user} leaving joinable group {group} due to auto_leave")
            request_info = user.username + ":" + group.name
            log = RequestLog(request_type=True, group=group, request_info=request_info, action=1, request_actor=user)
            log.save()
            user.groups.remove(group)
            ok = True
            status = GroupRequestLeaveStatus.LEFT
        else:
            grouprequest = GroupRequest()
            grouprequest.group = group
            grouprequest.user = user
            grouprequest.leave_request = True
            grouprequest.save()
            logger.info(f"Created group leave request for user {user} to group {group}")
            grouprequest.notify_leaders()

            status = GroupRequestLeaveStatus.CREATED_LEAVE_REQUEST
            ok = True

        return cls(ok=ok, status=status)


class GroupMembershipRemove(graphene.Mutation):
    class Arguments:
        group_id = graphene.Int(required=True)
        user_id = graphene.Int(required=True)

    ok = graphene.Boolean()
    error = graphene.String()

    @classmethod
    @login_required
    @user_passes_test(GroupManager.can_manage_groups)
    def mutate(cls, root, info, group_id, user_id):
        ok = False
        error = None
        a_user = info.context.user
        logger.debug(f"group_membership_remove called by user {a_user} for group id {group_id} on user id {user_id}")
        try:
            group = Group.objects.get(pk=group_id)
            # Check its a joinable group i.e. not corp or internal
            # And the user has permission to manage it
            if not GroupManager.check_internal_group(group) or not GroupManager.can_manage_group(a_user, group):
                logger.warning(f"User {a_user} attempted to remove a user from group {group_id} but permission was denied")
                error = 'Permission denied'
            else:
                try:
                    user = group.user_set.get(id=user_id)
                    request_info = user.username + ":" + group.name
                    log = RequestLog(request_type=None, group=group, request_info=request_info, action=1, request_actor=a_user)
                    log.save()
                    # Remove group from user
                    user.groups.remove(group)
                    logger.info(f"User {a_user} removed user {user} from group {group}")
                    ok = True
                except ObjectDoesNotExist:
                    error = "User does not exist in that group"

        except ObjectDoesNotExist:
            error = "Group does not exist"

        return cls(ok=ok, error=error)


class GroupMembershipAcceptRequest(graphene.Mutation):
    class Arguments:
        group_request_id = graphene.Int(required=True)

    ok = graphene.Boolean()
    error = graphene.String()

    @classmethod
    @login_required
    @user_passes_test(GroupManager.can_manage_groups)
    def mutate(cls, root, info, group_request_id):
        user = info.context.user
        ok = True
        error = None
        logger.debug(f"group_accept_request called by user {user} for grouprequest id {group_request_id}")
        try:
            group_request = (
                GroupRequest.objects
                .select_related('group', 'user__profile__state')
                .get(id=group_request_id)
            )
        except GroupRequest.DoesNotExist:
            error = "Group request doesn't exist"
            ok = False
        else:
            try:
                if not GroupManager.joinable_group(group_request.group, group_request.user.profile.state) or \
                        not GroupManager.can_manage_group(user, group_request.group):
                    logger.warning(f"User {user} attempted to accept group join request {group_request_id} but permission was denied")
                    error = "Permission denied"
                    ok = False
                else:
                    group_request.user.groups.add(group_request.group)
                    RequestLog.objects.create(
                        request_type=group_request.leave_request,
                        group=group_request.group,
                        request_info=str(group_request),
                        action=1,
                        request_actor=user
                    )

                    group_request.delete()
                    logger.info(f"User {user} accepted group request from user {group_request.user} to group {group_request.group.name}")
                    notify(
                        group_request.user,
                        "Group Application Accepted",
                        level="success",
                        message=f"Your application to {group_request.group} has been accepted."
                    )
            except:
                error = f'An unhandled error occurred while processing the application from {group_request.main_char} to {group_request.group}.'
                logger.exception(f"Unhandled exception occurred while user {user} attempting to accept grouprequest id {group_request_id}.")
                ok = False

        return cls(ok=ok, error=error)


class GroupMembershipRejectRequest(graphene.Mutation):
    class Arguments:
        group_request_id = graphene.Int(required=True)

    ok = graphene.Boolean()
    error = graphene.String()

    @classmethod
    @login_required
    @user_passes_test(GroupManager.can_manage_groups)
    def mutate(cls, root, info, group_request_id):
        user = info.context.user
        ok = True
        error = None

        logger.debug(f"group_reject_request called by user {user} for group request id {group_request_id}")
        try:
            group_request = (
                GroupRequest.objects
                .select_related('group', 'user')
                .get(id=group_request_id)
            )
        except GroupRequest.DoesNotExist:
            error = "Group request doesn't exist"
            ok = False
        else:
            try:
                if not GroupManager.can_manage_group(user, group_request.group):
                    logger.warning(f"User {user} attempted to reject group join request {group_request_id} but permission was denied")
                    error = "Permission denied"
                    ok = False
                else:
                    logger.info(f"User {user} rejected group request from user {group_request.user} to group {group_request.group.name}")
                    RequestLog.objects.create(
                        request_type=group_request.leave_request,
                        group=group_request.group,
                        request_info=str(group_request),
                        action=0,
                        request_actor=user
                    )
                    group_request.delete()
                    notify(
                        group_request.user,
                        "Group Application Rejected",
                        level="danger",
                        message=f"Your application to {group_request.group} has been rejected."
                    )
            except:
                error = f'An unhandled error occurred while processing the application from {group_request.main_char} to {group_request.group}.'
                logger.exception(f"Unhandled exception occurred while user {user} attempting to accept grouprequest id {group_request_id}.")
                ok = False

        return cls(ok=ok, error=error)


class GroupLeaveAcceptRequest(graphene.Mutation):
    class Arguments:
        group_request_id = graphene.Int(required=True)

    ok = graphene.Boolean()
    error = graphene.String()

    @classmethod
    @login_required
    @user_passes_test(GroupManager.can_manage_groups)
    def mutate(cls, root, info, group_request_id):
        user = info.context.user
        ok = True
        error = None
        logger.debug(f"group_leave_accept_request called by user {user} for group request id {group_request_id}")
        try:
            group_request = (
                GroupRequest.objects
                .select_related('group', 'user')
                .get(id=group_request_id)
            )
        except GroupRequest.DoesNotExist:
            error = "Group request doesn't exist"
            ok = False
        else:
            try:
                if not GroupManager.can_manage_group(user, group_request.group):
                    logger.warning(f"User {user} attempted to accept group leave request {group_request_id} but permission was denied")
                    error = "Permission denied"
                    ok = False
                else:
                    group_request.user.groups.remove(group_request.group)
                    RequestLog.objects.create(
                        request_type=group_request.leave_request,
                        group=group_request.group,
                        request_info=str(group_request),
                        action=1,
                        request_actor=user
                    )
                    group_request.delete()
                    logger.info(f"User {user} accepted group leave request from user {group_request.user} to group {group_request.group.name}")
                    notify(
                        group_request.user,
                        "Group Leave Request Accepted",
                        level="success",
                        message=f"Your request to leave {group_request.group} has been accepted."
                    )
            except:
                error = f'An unhandled error occurred while processing the application from {group_request.main_char} to {group_request.group}.'
                logger.exception(f"Unhandled exception occurred while user {user} attempting to accept grouprequest id {group_request_id}.")
                ok = False

        return cls(ok=ok, error=error)


class GroupLeaveRejectRequest(graphene.Mutation):
    class Arguments:
        group_request_id = graphene.Int(required=True)

    ok = graphene.Boolean()
    error = graphene.String()

    @classmethod
    @login_required
    @user_passes_test(GroupManager.can_manage_groups)
    def mutate(cls, root, info, group_request_id):
        user = info.context.user
        ok = True
        error = None

        logger.debug(f"group_leave_reject_request called by user {user} for group request id {group_request_id}")
        try:
            group_request = (
                GroupRequest.objects
                .select_related('group', 'user')
                .get(id=group_request_id)
            )
        except GroupRequest.DoesNotExist:
            error = "Group request doesn't exist"
            ok = False
        else:
            try:
                if not GroupManager.can_manage_group(user, group_request.group):
                    logger.warning(f"User {user} attempted to reject group leave request {group_request_id} but permission was denied")
                    error = "Permission denied"
                    ok = False
                else:
                    RequestLog.objects.create(
                        request_type=group_request.leave_request,
                        group=group_request.group,
                        request_info=str(group_request),
                        action=0,
                        request_actor=user
                    )
                    group_request.delete()
                    logger.info(f"User {user} rejected group leave request from user {group_request.user} for group {group_request.group.name}")
                    notify(
                        group_request.user,
                        "Group Leave Request Rejected",
                        level="danger",
                        message=f"Your request to leave {group_request.group} has been rejected."
                    )
            except:
                error = f'An unhandled error occurred while processing the application from {group_request.main_char} to {group_request.group}.'
                logger.exception(f"Unhandled exception occurred while user {user} attempting to accept grouprequest id {group_request_id}.")
                ok = False

        return cls(ok=ok, error=error)


class Mutation:
    groupmanagement_join_group_request = AddGroupRequest.Field()
    groupmanagement_leave_group_request = LeaveGroupRequest.Field()
    groupmanagement_remove_member = GroupMembershipRemove.Field()
    groupmanagement_accept_join_request = GroupMembershipAcceptRequest.Field()
    groupmanagement_reject_join_request = GroupMembershipRejectRequest.Field()
    groupmanagement_accept_leave_request = GroupLeaveAcceptRequest.Field()
    groupmanagement_reject_leave_request = GroupLeaveRejectRequest.Field()



import graphene
from graphql_jwt.decorators import login_required

from .types import TokenType


class Query:
    esi_user_tokens = graphene.List(TokenType)

    @login_required
    def resolve_esi_user_tokens(self, info):
        return info.context.user.token_set.all()



import datetime
from graphene_django.utils.testing import GraphQLTestCase
from unittest.mock import patch

from django.utils import timezone
from django.db.models import Min

from allianceauth.tests.auth_utils import AuthUtils
from app_utils.testdata_factories import UserFactory, UserMainFactory
from app_utils.esi_testing import EsiEndpoint, EsiClientStub
from app_utils.testing import generate_invalid_pk

from allianceauth.srp.models import SrpFleetMain, SrpUserRequest
from allianceauth.eveonline.models import EveCharacter
from allianceauth.notifications.models import Notification


class TestQueries(GraphQLTestCase):
    maxDiff = None

    @classmethod
    def setUpTestData(cls):
        cls.user = AuthUtils.add_permission_to_user_by_name('srp.access_srp', UserFactory(), False)

        cls.open_fleet = SrpFleetMain.objects.create(
            fleet_name='Test Fleet',
            fleet_doctrine='Test Doctrine',
            fleet_time=timezone.now() - datetime.timedelta(hours=1),
            fleet_srp_code='TEST',
            fleet_srp_status='',
        )

        cls.completed_fleet = SrpFleetMain.objects.create(
            fleet_name='Test Fleet 2',
            fleet_doctrine='Test Doctrine 2',
            fleet_time=timezone.now() - datetime.timedelta(hours=2),
            fleet_srp_code='TEST2',
            fleet_srp_status='Completed',
        )

    def test_srp_get_fleets_all(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            query($all: Boolean) {
                srpGetFleets(all: $all) {
                    id
                }
            }
            ''',
            variables={'all': True}
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'srpGetFleets': [
                        {
                            'id': str(self.open_fleet.pk),
                        },
                        {
                            'id': str(self.completed_fleet.pk),
                        },
                    ]
                }
            }
        )

    def test_srp_get_fleets_not_all(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            query($all: Boolean) {
                srpGetFleets(all: $all) {
                    id
                }
            }
            '''
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'srpGetFleets': [
                        {
                            'id': str(self.open_fleet.pk),
                        },
                    ]
                }
            }
        )


class TestAddFleetMutation(GraphQLTestCase):
    maxDiff = None

    @classmethod
    def setUpTestData(cls):
        cls.user = UserMainFactory()

        cls.input_data = {
            'fleetName': 'Test Fleet',
            'fleetDoctrine': 'Test Doctrine',
            'fleetTime': (timezone.now() - datetime.timedelta(hours=1)).isoformat(),
        }

    def test_decorator_no_perms(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            mutation($input: AddFleetMutationInput!) {
                srpAddFleet(input: $input) {
                    ok
                }
            }
            ''',
            input_data=self.input_data
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'srpAddFleet': None
                },
                'errors': [
                    {
                        'message': 'You do not have permission to perform this action',
                        'locations': [
                            {
                                'line': 3,
                                'column': 17
                            }
                        ],
                        'path': [
                            'srpAddFleet'
                        ]
                    }
                ]
            }
        )

    def test_ok(self):
        user = AuthUtils.add_permission_to_user_by_name('srp.add_srpfleetmain', self.user, False)

        self.client.force_login(user)

        response = self.query(
            '''
            mutation($input: AddFleetMutationInput!) {
                srpAddFleet(input: $input) {
                    ok
                    srpFleet {
                        fleetName
                        fleetDoctrine
                        fleetCommander {
                            id
                        }
                    }
                }
            }
            ''',
            input_data=self.input_data
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'srpAddFleet': {
                        'ok': True,
                        'srpFleet': {
                            'fleetName': self.input_data['fleetName'],
                            'fleetDoctrine': self.input_data['fleetDoctrine'],
                            'fleetCommander': {
                                'id': str(user.profile.main_character.pk),
                            },
                        }
                    }
                }
            }
        )

        self.assertEqual(SrpFleetMain.objects.count(), 1)


class TestRemoveDisableEnableCompletedUncompletedFleetMutation(GraphQLTestCase):
    maxDiff = None

    @classmethod
    def setUpTestData(cls):
        cls.user = AuthUtils.add_permission_to_user_by_name('auth.srp_management', UserFactory(), False)

        cls.fleet = SrpFleetMain.objects.create(
            fleet_name='Test Fleet',
            fleet_doctrine='Test Doctrine',
            fleet_time=timezone.now() - datetime.timedelta(hours=1),
            fleet_srp_code='TEST',
            fleet_srp_status='',
        )

    def test_remove_ok(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            mutation($fleetId: ID!) {
                srpRemoveFleet(fleetId: $fleetId) {
                    ok
                }
            }
            ''',
            variables={'fleetId': self.fleet.pk}
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'srpRemoveFleet': {
                        'ok': True,
                    }
                }
            }
        )

        self.assertEqual(SrpFleetMain.objects.count(), 0)

    def test_disable_ok(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            mutation($fleetId: ID!) {
                srpDisableFleet(fleetId: $fleetId) {
                    ok
                    srpFleet {
                        id
                        fleetSrpCode
                    }
                }
            }
            ''',
            variables={'fleetId': self.fleet.pk}
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'srpDisableFleet': {
                        'ok': True,
                        'srpFleet': {
                            'id': str(self.fleet.pk),
                            'fleetSrpCode': '',
                        }
                    }
                }
            }
        )

    @patch('allianceauth_graphql.srp.mutations.random_string')
    def test_enable_ok(self, mock_random_string):
        self.client.force_login(self.user)

        mock_random_string.return_value = 'TESTTEST'

        self.fleet.fleet_srp_code = ''
        self.fleet.save()

        response = self.query(
            '''
            mutation($fleetId: ID!) {
                srpEnableFleet(fleetId: $fleetId) {
                    ok
                    srpFleet {
                        id
                        fleetSrpCode
                    }
                }
            }
            ''',
            variables={'fleetId': self.fleet.pk}
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'srpEnableFleet': {
                        'ok': True,
                        'srpFleet': {
                            'id': str(self.fleet.pk),
                            'fleetSrpCode': 'TESTTEST',
                        }
                    }
                }
            }
        )

    def test_completed_ok(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            mutation($fleetId: ID!) {
                srpMarkCompleted(fleetId: $fleetId) {
                    ok
                    srpFleet {
                        id
                        fleetSrpStatus
                    }
                }
            }
            ''',
            variables={'fleetId': self.fleet.pk}
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'srpMarkCompleted': {
                        'ok': True,
                        'srpFleet': {
                            'id': str(self.fleet.pk),
                            'fleetSrpStatus': 'Completed',
                        }
                    }
                }
            }
        )

    def test_uncompleted_ok(self):
        self.client.force_login(self.user)

        self.fleet.fleet_srp_status = 'Completed'
        self.fleet.save()

        response = self.query(
            '''
            mutation($fleetId: ID!) {
                srpMarkUncompleted(fleetId: $fleetId) {
                    ok
                    srpFleet {
                        id
                        fleetSrpStatus
                    }
                }
            }
            ''',
            variables={'fleetId': self.fleet.pk}
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'srpMarkUncompleted': {
                        'ok': True,
                        'srpFleet': {
                            'id': str(self.fleet.pk),
                            'fleetSrpStatus': '',
                        }
                    }
                }
            }
        )


class TestSrpFleetUserRequestFormMutation(GraphQLTestCase):
    maxDiff = None

    @classmethod
    def setUpTestData(cls):
        cls.user = AuthUtils.add_permission_to_user_by_name('srp.access_srp', UserMainFactory(), False)

        cls.fleet = SrpFleetMain.objects.create(
            fleet_name='Test Fleet',
            fleet_doctrine='Test Doctrine',
            fleet_time=timezone.now() - datetime.timedelta(hours=1),
            fleet_srp_code='TEST',
            fleet_srp_status='',
        )

        cls.form_data = {
            'fleetSrpCode': 'TEST',
            'killboardLink': 'https://zkillboard.com/kill/1234567890/',
        }

    @patch('allianceauth.srp.managers.SRPManager.get_kill_data')
    @patch('allianceauth_graphql.srp.mutations.esi')
    def test_ok(self, mock_esi, mock_get_kill_data):
        self.client.force_login(self.user)

        endpoints = [
            EsiEndpoint(
                'Universe',
                'get_universe_types_type_id',
                'type_id',
                data={
                    '11567': {
                        'name': 'Avatar'
                    }
                }
            ),
        ]

        mock_esi.client = EsiClientStub.create_from_endpoints(endpoints)
        mock_get_kill_data.return_value = (11567, 64_840_457_150.95, self.user.profile.main_character.character_id)

        response = self.query(
            '''
            mutation($input: SrpFleetUserRequestFormMutationInput!) {
                srpRequest(input: $input) {
                    ok
                    srpRequest {
                        killboardLink
                        srpFleetMain {
                            id
                        }
                    }
                }
            }
            ''',
            input_data=self.form_data
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'srpRequest': {
                        'ok': True,
                        'srpRequest': {
                            'killboardLink': self.form_data['killboardLink'],
                            'srpFleetMain': {
                                'id': str(self.fleet.pk),
                            }
                        }
                    }
                }
            }
        )

        self.assertEqual(SrpUserRequest.objects.count(), 1)

    @patch('allianceauth.srp.managers.SRPManager.get_kill_data')
    def test_value_error(self, mock_get_kill_data):
        self.client.force_login(self.user)

        mock_get_kill_data.side_effect = ValueError('Test')

        response = self.query(
            '''
            mutation($input: SrpFleetUserRequestFormMutationInput!) {
                srpRequest(input: $input) {
                    ok
                    srpRequest {
                        killboardLink
                        srpFleetMain {
                            id
                        }
                    }
                }
            }
            ''',
            input_data=self.form_data
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'srpRequest': {
                        'ok': False,
                        'srpRequest': None
                    }
                }
            }
        )

        self.assertEqual(SrpUserRequest.objects.count(), 0)

    @patch('allianceauth.srp.managers.SRPManager.get_kill_data')
    def test_missing_character(self, mock_get_kill_data):
        self.client.force_login(self.user)

        mock_get_kill_data.return_value = (
            11567,
            64_840_457_150.95,
            EveCharacter.objects
            .aggregate(Min('character_id'))['character_id__min'] - 1
        )

        response = self.query(
            '''
            mutation($input: SrpFleetUserRequestFormMutationInput!) {
                srpRequest(input: $input) {
                    ok
                    srpRequest {
                        killboardLink
                        srpFleetMain {
                            id
                        }
                    }
                }
            }
            ''',
            input_data=self.form_data
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'srpRequest': {
                        'ok': False,
                        'srpRequest': None
                    }
                }
            }
        )

        self.assertEqual(SrpUserRequest.objects.count(), 0)

    def test_already_requested(self):
        self.client.force_login(self.user)

        SrpUserRequest.objects.create(
            srp_fleet_main=self.fleet,
            character=self.user.profile.main_character,
            killboard_link=self.form_data['killboardLink'],
            srp_ship_name='Avatar',
        )

        response = self.query(
            '''
            mutation($input: SrpFleetUserRequestFormMutationInput!) {
                srpRequest(input: $input) {
                    ok
                    srpRequest {
                        killboardLink
                        srpFleetMain {
                            id
                        }
                    }
                }
            }
            ''',
            input_data=self.form_data
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'srpRequest': {
                        'ok': False,
                        'srpRequest': None
                    }
                }
            }
        )

        self.assertEqual(SrpUserRequest.objects.count(), 1)


class TestSrpRequestRemoveMutation(GraphQLTestCase):
    maxDiff = None

    @classmethod
    def setUpTestData(cls):
        cls.user = AuthUtils.add_permission_to_user_by_name('auth.srp_management', UserMainFactory(), False)

        cls.fleet = SrpFleetMain.objects.create(
            fleet_name='Test Fleet',
            fleet_doctrine='Test Doctrine',
            fleet_time=timezone.now() - datetime.timedelta(hours=1),
            fleet_srp_code='TEST',
            fleet_srp_status='',
        )

        cls.request = SrpUserRequest.objects.create(
            srp_fleet_main=cls.fleet,
            character=cls.user.profile.main_character,
            killboard_link='https://zkillboard.com/kill/1234567890/',
            srp_ship_name='Avatar',
        )

    def test_ok(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            mutation($requestIds: [ID!]!) {
                srpRemoveRequests(requestIds: $requestIds) {
                    ok
                }
            }
            ''',
            variables={'requestIds': [self.request.pk]}
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'srpRemoveRequests': {
                        'ok': True,
                    }
                }
            }
        )

        self.assertEqual(SrpUserRequest.objects.count(), 0)


class TestSrpRequestApproveRejectMutation(GraphQLTestCase):
    maxDiff = None

    @classmethod
    def setUpTestData(cls):
        cls.user = AuthUtils.add_permission_to_user_by_name('auth.srp_management', UserMainFactory(), False)

        fleet = SrpFleetMain.objects.create(
            fleet_name='Test Fleet',
            fleet_doctrine='Test Doctrine',
            fleet_time=timezone.now() - datetime.timedelta(hours=1),
            fleet_srp_code='TEST',
            fleet_srp_status='',
        )

        cls.request = SrpUserRequest.objects.create(
            srp_fleet_main=fleet,
            character=cls.user.profile.main_character,
            killboard_link='https://zkillboard.com/kill/1234567890/',
            srp_ship_name='Avatar',
            srp_total_amount=64_840_457_150.95,
        )

        cls.request2 = SrpUserRequest.objects.create(
            srp_fleet_main=fleet,
            character=cls.user.profile.main_character,
            killboard_link='https://zkillboard.com/kill/12345067890/',
            srp_ship_name='Avatar',
            kb_total_loss=64_840_457_150.95,
        )

    def test_approve_ok(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            mutation($requestIds: [ID!]!) {
                srpApproveRequests(requestIds: $requestIds) {
                    ok
                }
            }
            ''',
            variables={
                'requestIds': [
                    self.request.pk,
                    self.request2.pk,
                    generate_invalid_pk(SrpUserRequest)
                ]
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'srpApproveRequests': {
                        'ok': True,
                    }
                }
            }
        )

        self.assertEqual(SrpUserRequest.objects.filter(srp_status='Approved').count(), 2)
        self.assertEqual(Notification.objects.count(), 2)

    def test_reject_ok(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            mutation($requestIds: [ID!]!) {
                srpRejectRequests(requestIds: $requestIds) {
                    ok
                }
            }
            ''',
            variables={
                'requestIds': [
                    self.request.pk,
                    self.request2.pk,
                    generate_invalid_pk(SrpUserRequest)
                ]
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'srpRejectRequests': {
                        'ok': True,
                    }
                }
            }
        )

        self.assertEqual(SrpUserRequest.objects.filter(srp_status='Rejected').count(), 2)
        self.assertEqual(Notification.objects.count(), 2)


class TestSrpUpdateAmountMutation(GraphQLTestCase):
    maxDiff = None

    @classmethod
    def setUpTestData(cls):
        cls.user = AuthUtils.add_permission_to_user_by_name('auth.srp_management', UserMainFactory(), False)

        fleet = SrpFleetMain.objects.create(
            fleet_name='Test Fleet',
            fleet_doctrine='Test Doctrine',
            fleet_time=timezone.now() - datetime.timedelta(hours=1),
            fleet_srp_code='TEST',
            fleet_srp_status='',
        )

        cls.request = SrpUserRequest.objects.create(
            srp_fleet_main=fleet,
            character=cls.user.profile.main_character,
            killboard_link='https://zkillboard.com/kill/1234567890/',
            srp_ship_name='Avatar',
            srp_total_amount=64_840_457_150.95,
        )

    def test_ok(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            mutation($fleetSrpRequestId: ID!, $srpTotalAmount: Float!) {
                srpUpdateAmount(fleetSrpRequestId: $fleetSrpRequestId, srpTotalAmount: $srpTotalAmount) {
                    ok
                }
            }
            ''',
            variables={
                'fleetSrpRequestId': self.request.pk,
                'srpTotalAmount': 1234.56
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'srpUpdateAmount': {
                        'ok': True,
                    }
                }
            }
        )

        self.request.refresh_from_db()

        self.assertEqual(self.request.srp_total_amount, 1234)

    def test_missing_request(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            mutation($fleetSrpRequestId: ID!, $srpTotalAmount: Float!) {
                srpUpdateAmount(fleetSrpRequestId: $fleetSrpRequestId, srpTotalAmount: $srpTotalAmount) {
                    ok
                }
            }
            ''',
            variables={
                'fleetSrpRequestId': generate_invalid_pk(SrpUserRequest),
                'srpTotalAmount': 1234.56
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'srpUpdateAmount': {
                        'ok': False,
                    }
                }
            }
        )

        self.request.refresh_from_db()

        self.assertEqual(self.request.srp_total_amount, 64_840_457_150)


class TestSrpUpdateAARMutation(GraphQLTestCase):
    maxDiff = None

    @classmethod
    def setUpTestData(cls):
        cls.user = AuthUtils.add_permission_to_user_by_name('auth.srp_management', UserFactory(), False)

        cls.fleet = SrpFleetMain.objects.create(
            fleet_name='Test Fleet',
            fleet_doctrine='Test Doctrine',
            fleet_time=timezone.now() - datetime.timedelta(hours=1),
            fleet_srp_code='TEST',
            fleet_srp_status='',
        )

    def test_ok(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            mutation($fleetId: ID!, $fleetSrpAarLink: String!) {
                srpUpdateAar(fleetId: $fleetId, fleetSrpAarLink: $fleetSrpAarLink) {
                    ok
                }
            }
            ''',
            variables={
                'fleetId': self.fleet.pk,
                'fleetSrpAarLink': 'Test AAR'
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'srpUpdateAar': {
                        'ok': True,
                    }
                }
            }
        )

        self.fleet.refresh_from_db()

        self.assertEqual(self.fleet.fleet_srp_aar_link, 'Test AAR')



import graphene
from graphene_django import DjangoObjectType

from allianceauth.srp.models import SrpFleetMain, SrpUserRequest


class SrpFleetMainType(DjangoObjectType):
    total_cost = graphene.Int(required=True)
    pending_requests = graphene.Int(required=True)

    class Meta:
        model = SrpFleetMain


class SrpUserRequestType(DjangoObjectType):
    class Meta:
        model = SrpUserRequest



import graphene
from graphql_jwt.decorators import login_required, user_passes_test, permission_required

from allianceauth.corputils.views import access_corpstats_test, SWAGGER_SPEC_PATH
from allianceauth.corputils.models import CorpStats
from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo

from ..decorators import tokens_required

from esi.models import Token


class AddCorpStatsMutation(graphene.Mutation):
    class Arguments:
        token_id = graphene.ID(required=True)

    ok = graphene.Boolean()

    @classmethod
    @login_required
    @user_passes_test(access_corpstats_test)
    @permission_required('corputils.add_corpstats')
    @tokens_required(scopes='esi-corporations.read_corporation_membership.v1')
    def mutate(cls, root, info, token_id):
        user = info.context.user
        token = Token.objects.get(pk=token_id)
        if token.user != user:
            raise PermissionError("Token not valid")

        if EveCharacter.objects.filter(character_id=token.character_id).exists():
            corp_id = EveCharacter.objects.get(character_id=token.character_id).corporation_id
        else:
            corp_id = \
                token.get_esi_client(spec_file=SWAGGER_SPEC_PATH).Character.get_characters_character_id(
                    character_id=token.character_id).result()['corporation_id']
        try:
            corp = EveCorporationInfo.objects.get(corporation_id=corp_id)
        except EveCorporationInfo.DoesNotExist:
            corp = EveCorporationInfo.objects.create_corporation(corp_id)
        cs = CorpStats.objects.create(token=token, corp=corp)
        cs.update()
        assert cs.pk  # ensure update was successful

        return cls(ok=True)


class UpdateCorpStatsMutation(graphene.Mutation):
    class Arguments:
        corp_id = graphene.Int(required=True)

    ok = graphene.Boolean()

    @classmethod
    @login_required
    @user_passes_test(access_corpstats_test)
    def mutate(cls, root, info, corp_id):
        corp = EveCorporationInfo.objects.get(corporation_id=corp_id)
        corpstats = CorpStats.objects.get(corp=corp)
        corpstats.update()
        return cls(ok=True)


class Mutation:
    corputils_add_corpstats = AddCorpStatsMutation.Field()
    corputils_update_corpstats = UpdateCorpStatsMutation.Field()
